﻿using System;
using System.Collections.Generic;
using com.middlemind.Eds3;
using com.middlemind.Eds3.bhw;
using com.middlemind.Eds3.mock;
using com.middlemind.Eds3.unity;
using System.Threading;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3App
{
    public class Eds3RunApp
    {
        public static int MAX_CLIENTS = 6;
        public static List<Eds3Client> CLIENTS = new List<Eds3Client>();

        /**
         * @param args the command line arguments
         */
        public static void Main(String[] args)
        {
            try
            {
                if (args != null)
                {
                    Utils.wr("Found " + args.Length + " args.");
                    Utils.wr("Found working dir: " + System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location));
                }

                if (args != null && args.Length >= 1)
                {
                    Eds3Server eds3server = null;
                    Eds3Client eds3client = null;
                    Eds3ServerBhw eds3serverBhw = null;

                    if (args[0].ToLower().Equals("server"))
                    {
                        eds3server = new Eds3Server();

                        if (args.Length >= 2)
                        {
                            eds3server.SetLocalIpAddress(args[1]);
                        }

                        if (args.Length >= 3)
                        {
                            eds3server.SetPort(args[2]);
                        }

                        bool tmp = false;
                        if(args.Length >= 4)
                        {
                            tmp = bool.Parse(args[3]);
                            Utils.LOGGING_ON = tmp;
                            eds3server.SetDebugOn(tmp);
                        }

                        if (args.Length >= 5)
                        {
                            Utils.VERBOSE = bool.Parse(args[4]);
                        }

                        GameRoom.PLAYER_THRESHOLD_INACTIVE_COMM_MS = 3000;
                        eds3server.Start();

                        String inp = "";
                        System.Console.Out.WriteLine("Enter 'q' to exit.");
                        while (inp != null && !inp.Equals("q"))
                        {
                            System.Console.Out.Write("Entry: ");
                            inp = System.Console.In.ReadLine().ToLower().Trim();
                            Utils.wr("Found cmd: " + inp);
                            if (!inp.Equals("q"))
                            {
                                eds3server.ProcessCmd(inp);
                                Thread.Sleep(2000);
                            }
                        }

                        Utils.wr("Out of loop...");
                        eds3server.Stop();
                        Utils.wr("Exiting...");
                        System.Environment.Exit(1);
                        return;

                    }
                    else if (args[0].ToLower().Equals("bhw_server"))
                    {
                        eds3serverBhw = new Eds3ServerBhw();

                        if (args.Length >= 2)
                        {
                            eds3serverBhw.SetLocalIpAddress(args[1]);
                        }

                        if (args.Length >= 3)
                        {
                            eds3serverBhw.SetPort(args[2]);
                        }

                        bool tmp = false;
                        if (args.Length >= 4)
                        {
                            tmp = bool.Parse(args[3]);
                            Utils.LOGGING_ON = tmp;
                            eds3server.SetDebugOn(tmp);
                        }

                        if (args.Length >= 5)
                        {
                            Utils.VERBOSE = bool.Parse(args[4]);
                        }

                        GameRoom.PLAYER_THRESHOLD_INACTIVE_COMM_MS = 3000;
                        eds3serverBhw.Start();

                        String inp = "";
                        System.Console.Out.WriteLine("Enter 'q' to exit.");
                        while (inp != null && !inp.Equals("q"))
                        {
                            System.Console.Out.Write("Entry: ");
                            inp = System.Console.In.ReadLine().ToLower().Trim();
                            Utils.wr("Found cmd: " + inp);
                            if (!inp.Equals("q"))
                            {
                                eds3serverBhw.ProcessCmd(inp);
                                Thread.Sleep(2000);
                            }
                        }

                        Utils.wr("Out of loop...");
                        eds3serverBhw.Stop();
                        Utils.wr("Exiting...");
                        System.Environment.Exit(1);
                        return;

                    }
                    else if (args[0].ToLower().Equals("client"))
                    {
                        Eds3Client.CONSOLIDATE_LOGS_ON = false;
                        eds3client = new Eds3Client(Eds3Server.GetLocalIPAddressStat(), 49999);

                        if (args.Length >= 2)
                        {
                            eds3client.SetLocalIpAddress(args[1]);
                        }

                        if (args.Length >= 3)
                        {
                            eds3client.SetPort(args[2]);
                        }

                        Eds3Client.NOT_UNITY = true;
                        eds3client.Start();

                        String inp = "";
                        System.Console.Out.WriteLine("Enter 'q' to exit.");
                        while (inp != null && !inp.Equals("q"))
                        {
                            System.Console.Out.Write("Entry: ");
                            inp = System.Console.In.ReadLine().ToLower().Trim();
                            Utils.wr("Found cmd: " + inp);
                            if (!inp.Equals("q"))
                            {
                                eds3client.ProcessCmd(inp);
                                Thread.Sleep(2000);
                            }
                        }

                        eds3client.Stop();
                        Utils.wr("Exiting...");
                        System.Environment.Exit(1);
                        return;

                    }
                    else if (args[0].ToLower().Equals("clients"))
                    {
                        CLIENTS.Clear();

                        Utils.wr("Starting " + MAX_CLIENTS + " new clients.");
                        Eds3Client tmpEds3client = null;
                        Eds3Client.CONSOLIDATE_LOGS_ON = false;
                        Eds3Client.NOT_UNITY = true;
                        for (int i = 0; i < MAX_CLIENTS; i++)
                        {
                            if (args.Length >= 2)
                            {
                                tmpEds3client = new Eds3Client(args[1], 49999);
                            }
                            else
                            {
                                tmpEds3client = new Eds3Client(Eds3Server.GetLocalIPAddressStat(), 49999);
                            }

                            if (args.Length >= 3)
                            {
                                tmpEds3client.SetPort(args[2]);
                            }
            
                            CLIENTS.Add(tmpEds3client);
                            tmpEds3client.Start();
                        }

                        String inp = "";
                        while (inp != null && !inp.Equals("q"))
                        {
                            System.Console.Out.Write("Entry: ");
                            inp = System.Console.In.ReadLine().ToLower();
                            Thread.Sleep(2000);
                        }

                        Utils.wr("Found stop command.");
                        Eds3Client eds3c;
                        for (int i = 0; i < CLIENTS.Count; i++)
                        {
                            eds3c = CLIENTS[i];
                            Utils.wr("Stopping client: " + eds3c.GetClientId() + " with SpawnId/PlayerId: " + eds3c.GetSpawnId());
                            eds3c.Stop();

                        }
                        Utils.wr("Exiting...");
                        System.Environment.Exit(1);
                        return;

                    }
                    else if (args[0].ToLower().Equals("sertest"))
                    {
                        MockGenDungeon.SLEEP = false;
                        MockGenDungeon mgd1 = new MockGenDungeon(100);
                        Utils.wr("Generating...");
                        long st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        mgd1.GenerateObjects();
                        mgd1.GenerateSpawnObjects();
                        long sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        Utils.wr("Generation Time MS: " + (sp - st));

                        Utils.wr("Serializing...");
                        st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        byte[] objs = mgd1.SerializeNetworkGameObjects();
                        sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        Utils.wr("Serialization Time MS: " + (sp - st));
                        Utils.wr("Found " + objs.Length + " bytes of serialized data.");
                        Utils.wr("Expected " + (GameObjectWrapperDefault.TOTAL_LENGTH * mgd1.GetNetworkGameObjectsCount()) + " bytes of serialized data.");

                        Utils.wr("Deserializing...");
                        st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        MockGenDungeon mgd2 = new MockGenDungeon(100);
                        mgd2.DeserializeNetworkGameObjects(objs);
                        sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        Utils.wr("Deserialization Time MS: " + (sp - st));

                        bool res = mgd1.NetworkGameObjectsEquals(mgd2);
                        Utils.wr("Serialization Test Success: " + res);
                        System.Environment.Exit(1);
                        return;

                    }
                    else if (args[0].ToLower().Equals("serstats"))
                    {
                        long totalGenTimeMs = 0;
                        long totalSerTimeMs = 0;
                        long totalDserTimeMs = 0;
                        int cnt = 10000;

                        long st;
                        long sp;
                        byte[] objs;
                        bool res;
                        MockGenDungeon mgd1;
                        MockGenDungeon mgd2;

                        Utils.wr("Running " + cnt + " tests...");
                        Utils.LOGGING_ON = false;
                        Utils.VERBOSE = false;
                        MockGenDungeon.SLEEP = false;

                        for (int i = 0; i < cnt; i++)
                        {

                            mgd1 = new MockGenDungeon(100);
                            st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            mgd1.GenerateObjects();
                            mgd1.GenerateSpawnObjects();
                            sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            totalGenTimeMs += (sp - st);

                            st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            objs = mgd1.SerializeNetworkGameObjects();
                            sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            totalSerTimeMs += (sp - st);

                            mgd2 = new MockGenDungeon(100);
                            st = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            mgd2.DeserializeNetworkGameObjects(objs);
                            sp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            totalDserTimeMs += (sp - st);

                            res = mgd1.NetworkGameObjectsEquals(mgd2);
                            if (!res)
                            {
                                Utils.LOGGING_ON = true;
                                Utils.wr("Comparison failure, iteration: " + i);
                                break;
                            }

                            mgd1.RandomlyAdjustObjects();
                        }

                        Utils.LOGGING_ON = true;
                        double avgGenTimeMs = (double)totalGenTimeMs / cnt;
                        double avgSerTimeMs = (double)totalSerTimeMs / cnt;
                        double avgDserTimeMs = (double)totalDserTimeMs / cnt;

                        Utils.wr("TotalGenTimeMs: " + totalGenTimeMs);
                        Utils.wr("TotalSerTimeMs: " + totalSerTimeMs);
                        Utils.wr("TotalDserTimeMs: " + totalDserTimeMs);

                        Utils.wr("AvgGenTimeMs: " + avgGenTimeMs);
                        Utils.wr("AvgSerTimeMs: " + avgSerTimeMs);
                        Utils.wr("AvgDserTimeMs: " + avgDserTimeMs);
                        Utils.wr("60 FPS = 1 frame every " + (1000 / 60) + " ms.");
                        System.Environment.Exit(1);
                        return;

                    }
                }
            }
            catch (Exception e)
            {
                Utils.wr(e.Message);
                Utils.wr(e.StackTrace);
            }
        }
    }
}
