﻿using System;
using System.Threading;
using System.Net;
using com.middlemind.Eds3.mock;
using System.Net.Sockets;
using com.middlemind.Eds3.unity;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3
{
    public class Eds3Client
    {
        #region enumerations
        public enum Eds3ClientMode
        {
            CLIENT_SEND_REQUEST_DUNGEN_SEED,
            CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED,
            CLIENT_SEND_REQUEST_SPAWN_ID,
            CLIENT_RECEIVE_REQUESTED_SPAWN_ID,
            IN_LOBBY,
            WAITING,
            PREV_MODE_EMPTY,
            MODE_EMPTY,
            CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1,
            CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1,
            RESTING,
            CLIENT_SEND_REQUEST_DROP_CLIENT,
            CLIENT_RECEIVE_REQUESTED_DROP_CLIENT,
            CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2,
            CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2,
            GENERAL_SEND_REQUEST_1,
            GENERAL_RECEIVE_REQUESTED_1,
            GENERAL_SEND_REQUEST_2,
            GENERAL_RECEIVE_REQUESTED_2,
            GENERAL_SEND_REQUEST_3,
            GENERAL_RECEIVE_REQUESTED_3,
        }

        public enum Eds3ClientObjSerMode
        {
            L1,
            L2
        }
        #endregion

        #region static_vars
        //STATIC VARIABLES
        public const int MAX_GLOBAL_ERRORS = 25;
        public const int MAX_ERRORS = 25;
        public const int MAX_MOCK_MESSAGES_OUT = 25;
        public const int DEFAULT_IN_BUFFER = 9000;
        public const int DEFAULT_OUT_BUFFER = 9000;
        public static bool MOCK_MODE_ON = true;
        public static bool ASYNC_ON = false;
        public static bool NOT_UNITY = false;
        private static Random rand = new System.Random((int)System.DateTime.Now.Ticks);
        private static bool REVERSE_TO_BYTES = false;
        private static bool REVERSE_FROM_BYTES = false;
        public static int TIMING_MS = 80;
        public static bool TIMING_MS_ON = true;
        private static int CLIENT_ID = 0;
        public static bool CONSOLIDATE_LOGS_ON = true;
        public static int CONSOLIDATE_LOGS_LEN = 128;
        public static Eds3ClientObjSerMode OBJ_SER_MODE = Eds3ClientObjSerMode.L2;
        public static int SO_TIMEOUT_MS = 80;
        public static bool FORCE_FLUSH = false;
        public static bool SERVER_DRIVEN_MODE_ON = true;
        public static bool WAIT_FOR_DROP_CLIENT_RESPONSE = false;
        #endregion

        #region local_vars
        //LOCAL VARIABLES
        private int totalDataIn = 0;
        private int totalMsgsIn = 0;
        private int totalDataOut = 0;
        private int totalMsgsOut = 0;

        private bool debugOn = true;
        private Socket clientSocket = null;
        private IPEndPoint endPoint = null;
        private EndPoint clientEndPoint = null;
        private EndPoint serverEndPoint = null;

        private int port = -1;
        private string host = "";
        private Eds3ClientMode mode;
        private Eds3ClientMode prevMode;
        private Eds3ClientMode rootMode;
        private string title = "Eds3Client";
        private bool shuttingDown = false;
        private string version = "0101";
        private string serverVersion = null;
        private int inBufferSize = DEFAULT_IN_BUFFER;
        private int outBufferSize = DEFAULT_OUT_BUFFER;
        private int dunGenSeed = -1;
        private int spawnId = -1;
        private MockGenDungeon mgd = null;
        private int errorCount = 0;
        private int clientId = 0;

        private Socket ancClientSocket;
        private IPEndPoint ancClientEndPoint;
        private EndPoint ancServerEndPoint;
        private int ancLen;
        private byte[] ancDataOut;
        private byte[] ancDataIn;
        private HandleEds3ClientComm commH = null;
        private bool forceNextMode = false;
        private Eds3ClientMode nextMode = Eds3ClientMode.IN_LOBBY;
        private HandleDebugOut debugOut = null;
        private int globalErrorCount = 0;
        private HandleGameObjectManager gomH = null;
        private bool paused = false;
        private string consolidatedLogs = "";
        private bool startShuttingDown = false;
        private GameObjectWrapper gowStatic = null;
        private long sync1Sent = 0;
        private long sync1Recv = 0;
        private long sync1Diff = 0;
        private long sync1Tmp = 0;
        private long sync2Sent = 0;
        private long sync2Recv = 0;
        private long sync2Diff = 0;
        private long sync2Tmp = 0;
        private long syncFinalDiff = 0;

        private Eds3ClientMode[] vals = new Eds3Client.Eds3ClientMode[]
        {
            Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED,
            Eds3Client.Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED,
            Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID,
            Eds3Client.Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_SPAWN_ID,
            Eds3Client.Eds3ClientMode.IN_LOBBY,
            Eds3Client.Eds3ClientMode.WAITING,
            Eds3Client.Eds3ClientMode.PREV_MODE_EMPTY,
            Eds3Client.Eds3ClientMode.MODE_EMPTY,
            Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1,
            Eds3Client.Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1,
            Eds3Client.Eds3ClientMode.RESTING,
            Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT,
            Eds3Client.Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DROP_CLIENT,
            Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2,
            Eds3Client.Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2,
            Eds3Client.Eds3ClientMode.GENERAL_SEND_REQUEST_1,
            Eds3Client.Eds3ClientMode.GENERAL_RECEIVE_REQUESTED_1,
            Eds3Client.Eds3ClientMode.GENERAL_SEND_REQUEST_2,
            Eds3Client.Eds3ClientMode.GENERAL_RECEIVE_REQUESTED_2,
            Eds3Client.Eds3ClientMode.GENERAL_SEND_REQUEST_3,
            Eds3Client.Eds3ClientMode.GENERAL_RECEIVE_REQUESTED_3,
        };
        #endregion

        #region constructors
        //CONSTRUCTORS
        public Eds3Client(string Host, int Port)
        {
            this.PrintEds3ClientModeValues();
            CLIENT_ID++;
            this.clientId = CLIENT_ID;
            this.host = Host;
            this.port = Port;
            this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
            this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
            this.gowStatic = this.GetEmptyGameObjectWrapper();
            System.Console.Out.WriteLine(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
            System.Console.Out.WriteLine(this.clientId + ": Version: " + this.version);
        }

        public Eds3Client(string Host, int Port, string Title)
        {
            this.PrintEds3ClientModeValues();
            CLIENT_ID++;
            this.clientId = CLIENT_ID;
            this.host = Host;
            this.port = Port;
            this.title = Title;
            this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
            this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
            this.gowStatic = this.GetEmptyGameObjectWrapper();
            System.Console.Out.WriteLine(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
            System.Console.Out.WriteLine(this.clientId + ": Version: " + this.version);
        }

        public Eds3Client(string Host, int Port, string Title, HandleEds3ClientComm CommH)
        {
            this.PrintEds3ClientModeValues();
            CLIENT_ID++;
            this.clientId = CLIENT_ID;
            this.host = Host;
            this.port = Port;
            this.title = Title;
            this.commH = CommH;
            this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
            this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
            this.gowStatic = this.GetEmptyGameObjectWrapper();

            if (this.commH != null)
            {
                this.commH.HandleEds3ClientInit(this);
            }

            System.Console.Out.WriteLine(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
            System.Console.Out.WriteLine(this.clientId + ": Version: " + this.version);
        }

        public Eds3Client(string Host, int Port, string Title, HandleEds3ClientComm CommH, HandleGameObjectManager GomH)
        {
            this.PrintEds3ClientModeValues();
            CLIENT_ID++;
            this.clientId = CLIENT_ID;
            this.host = Host;
            this.port = Port;
            this.title = Title;
            this.commH = CommH;
            this.gomH = GomH;
            this.mode = Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED;
            this.prevMode = Eds3ClientMode.PREV_MODE_EMPTY;
            this.gowStatic = this.GetEmptyGameObjectWrapper();

            if (this.commH != null)
            {
                this.commH.HandleEds3ClientInit(this);
            }

            System.Console.Out.WriteLine(this.clientId + ": Host:" + this.host + " Port: " + this.port + " Title: " + this.title);
            System.Console.Out.WriteLine(this.clientId + ": Version: " + this.version);
        }
        #endregion

        #region getters_and_setters
        public void SetLocalIpAddress(String ip)
        {
            this.host = ip;
        }

        public long GetSyncFinalDiff()
        {
            return this.syncFinalDiff;
        }
    
        public int GetPort()
        {
            return port;
        }

        public void SetPort(int p)
        {
            this.port = p;
        }

        public void SetPort(String p)
        {
            int tmp = -1;
            bool b = int.TryParse(p, out tmp);
            if (b)
            {
                this.port = tmp;
            }
        }

        public bool IsPaused()
        {
            return paused;
        }

        public void SetPaused(bool paused)
        {
            this.paused = paused;
        }

        public bool IsDebugOn()
        {
            return this.debugOn;
        }

        public void SetDebugOn(bool debugOn)
        {
            this.debugOn = debugOn;
        }

        public string GetServerVersion()
        {
            return this.serverVersion;
        }

        public int GetClientId()
        {
            return this.clientId;
        }

        public int GetSpawnId()
        {
            return this.spawnId;
        }

        public int GetDunGenSeed()
        {
            return this.dunGenSeed;
        }

        public string GetVersion()
        {
            return this.version;
        }

        public HandleEds3ClientComm GetCommHandler()
        {
            return this.commH;
        }

        public void SetCommHandler(HandleEds3ClientComm CommH)
        {
            this.commH = CommH;
        }

        public Eds3ClientMode GetMode()
        {
            return this.mode;
        }

        public void SetMode(Eds3ClientMode mode)
        {
            if (this.forceNextMode)
            {
                this.forceNextMode = false;
                this.prevMode = this.mode;
                this.mode = this.nextMode;

            }
            else
            {
                this.prevMode = this.mode;
                this.mode = mode;

            }
        }

        public HandleDebugOut GetHandleDebugOut()
        {
            return this.debugOut;
        }

        public void SetHandleDebugOut(HandleDebugOut dOut)
        {
            this.debugOut = dOut;
        }

        public int GetGlobalErrorCount()
        {
            return this.globalErrorCount;
        }

        public void SetGlobalErrorCount(int gec)
        {
            this.globalErrorCount = gec;
        }

        public HandleGameObjectManager GetHandleGameObjectManager()
        {
            return this.gomH;
        }

        public void SetHandleGameObjectManager(HandleGameObjectManager gomH)
        {
            this.gomH = gomH;
        }

        public void SetForceNextMode(Eds3ClientMode nMode)
        {
            this.forceNextMode = true;
            this.nextMode = nMode;
        }

        public bool IsShuttingDown()
        {
            return this.shuttingDown;
        }
        #endregion

        #region utility_methods
        public void PrintEds3ClientModeValues()
        {
            int len = this.vals.Length;

            for (int i = 0; i < len; i++)
            {
                wr("Idx: " + i + " Eds3ClientMode Value: " + (int)vals[i]);
            }
        }

        public GameObjectManager GetGameObjectManager()
        {
            if (this.gomH != null)
            {
                return this.gomH.GetGameObjectManager();
            }
            else
            {
                return this.mgd.GetGameObjectManager();
            }
        }

        public GameObjectWrapper GetPlayerChar()
        {
            if (this.gomH != null)
            {
                return this.gomH.GetPlayerChar();
            }
            else
            {
                return this.mgd.GetPlayerChar();
            }
        }

        public GameObjectWrapper GetEmptyGameObjectWrapper()
        {
            if (this.gomH != null)
            {
                return this.gomH.GetEmptyGameObjectWrapper();
            }
            else
            {
                return new GameObjectWrapperDefault();
            }
        }

        public void ProcessCmd(string cmd)
        {
            wr("client processCmd: " + cmd);
            if (cmd != null)
            {
                cmd = cmd.ToLower();

                if (cmd.IndexOf("netstats") != -1)
                {
                    wr(this.clientId + ": TotalNetworkDataIn: " + this.totalDataIn);
                    wr(this.clientId + ": TotalNetworkMsgsIn: " + this.totalMsgsIn);
                    wr(this.clientId + ": TotalNetworkDataOut: " + this.totalDataOut);
                    wr(this.clientId + ": TotalNetworkMsgsOut: " + this.totalMsgsOut);

                }

                if (cmd.IndexOf("safeclose") != -1)
                {
                    this.forceNextMode = true;
                    this.nextMode = Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT;

                }
            }
        }

        public void wr(string s)
        {
            if (s == null)
            {
                s = "";
            }

            if (CONSOLIDATE_LOGS_ON)
            {
                s += "\n";
                this.consolidatedLogs += s;
                if (this.consolidatedLogs.Length >= CONSOLIDATE_LOGS_LEN)
                {
                    PrivateWr(this.consolidatedLogs);
                    this.consolidatedLogs = "";
                }
            }
            else
            {
                PrivateWr(s);
            }
        }

        private void PrivateWr(string s)
        {
            if (s == null)
            {
                s = "";
            }

            if (this.debugOn == true)
            {
                if (this.debugOut != null)
                {
                    this.debugOut.wr(s);
                }
                else
                {
                    if (NOT_UNITY == true)
                    {
                        System.Console.Out.WriteLine(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() + ": " + s);
                        if (FORCE_FLUSH)
                        {
                            System.Console.Out.Flush();
                        }

                        System.Diagnostics.Debug.WriteLine(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() + ": " + s);
                        if (FORCE_FLUSH)
                        {
                            System.Diagnostics.Debug.Flush();
                        }
                    }
                }
            }
        }

        public void ForceWr(string s)
        {
            this.PrivateWr(s);
        }

        public void PrintStackTrace(string[] st)
        {
            if (st != null)
            {
                int len = st.Length;
                for (int i = 0; i < len; i++)
                {
                    wr(st[i].ToString());
                }
            }
        }

        private void PrintStackTrace(string st)
        {
            if (st != null)
            {
                wr(st);
            }
        }
        #endregion

        #region start_and_stop_methods
        public void Start()
        {
            wr(this.clientId + ": Starting: " + this.title + " Version: " + this.version);
            ThreadStart ts = new ThreadStart(StartWorking);
            Thread trd = new Thread(ts);
            trd.Start();
        }

        public void Stop()
        {
            this.shuttingDown = true;
            if (CONSOLIDATE_LOGS_ON)
            {
                this.ForceWr(this.consolidatedLogs);
            }

            try
            {
                this.clientSocket.Close(1);
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
            }
        }

        public void StartWorking()
        {
            try
            {
                wr(this.clientId + ": StartWorking");
                this.endPoint = new IPEndPoint(IPAddress.Parse(this.host), this.port);
                this.serverEndPoint = (EndPoint)this.endPoint;

                this.clientSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.IP);
                this.clientSocket.SendTimeout = SO_TIMEOUT_MS;
                this.clientSocket.ReceiveTimeout = SO_TIMEOUT_MS;
                wr(this.clientId + ": Working on port: " + this.port + " Local IP: " + this.endPoint.Address);

                this.shuttingDown = false;
                this.startShuttingDown = false;
                this.paused = false;
            
                while (!this.shuttingDown)
                {
                    try
                    {
                        if (TIMING_MS_ON)
                        {
                            Thread.Sleep(TIMING_MS);
                        }

                        if (this.paused)
                        {
                            Thread.Sleep(TIMING_MS);
                        }
                        else
                        {
                            if (!ASYNC_ON)
                            {
                                lock (this)
                                {
                                    this.ancClientSocket = this.clientSocket;
                                    this.ancClientEndPoint = this.endPoint;
                                    this.ancServerEndPoint = this.serverEndPoint;
                                }

                                ProcessCurrentMode();
                            }
                            else
                            {
                                AsyncProcessCurrentModeCallBack(this.clientSocket, this.serverEndPoint, this.endPoint);
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        if (!this.shuttingDown)
                        {
                            wr(e.Message);
                            PrintStackTrace(e.StackTrace);
                        }

                        this.forceNextMode = true;
                        this.nextMode = rootMode;
                        SetMode(rootMode);
                    }

                    if (this.shuttingDown)
                    {
                        break;
                    }
                }

            }
            catch (Exception e)
            {
                if (!this.shuttingDown)
                {
                    wr(e.Message);
                    PrintStackTrace(e.StackTrace);
                }
            }
        }

        public void run()
        {
            StartWorking();
        }
        #endregion

        #region network_methods
        public void AsyncProcessCurrentModeCallBack(Socket clientSocket, EndPoint serverEndPoint, IPEndPoint clientEndPoint)
        {
            lock (this)
            {
                this.ancClientSocket = clientSocket;
                this.ancServerEndPoint = serverEndPoint;
                this.ancClientEndPoint = clientEndPoint;
            }

            ThreadStart ts = new ThreadStart(ProcessCurrentMode);
            Thread trd = new Thread(ts);
            trd.Start();
        }

        public void ProcessCurrentMode()
        {
            try
            {
                Socket clientSocket = null;
                IPEndPoint clientEndPoint = null;
                EndPoint serverEndPoint = null;

                lock (this)
                {
                    clientSocket = this.ancClientSocket;
                    clientEndPoint = this.ancClientEndPoint;
                    serverEndPoint = this.ancServerEndPoint;
                }

                if (MOCK_MODE_ON)
                {
                    long st = (long)(rand.Next(3000) + 1);
                    wr("Sleep for " + st + " ms.");
                    Thread.Sleep((int)st);
                }

                if (this.paused)
                {
                    Thread.Sleep(TIMING_MS);
                    return;
                }

                if (this.shuttingDown)
                {
                    return;
                }

                if (this.startShuttingDown && this.GetMode() != Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)
                {
                    this.forceNextMode = true;
                    this.nextMode = Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT;
                    this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                }

                byte[] receiveData = new byte[inBufferSize];
                IPAddress ipAddress = clientEndPoint.Address;
                string ip = ipAddress.ToString();
                Eds3ClientMode lmode = this.GetMode();

                if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED)
                {
                    rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_DUNGEN_SEED");
                    wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_DUNGEN_SEED msg.");
                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DUNGEN_SEED, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + sentOn.Length;
                    sync1Sent = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length), sentOn.Length);

                    NetworkWriteCallBack(dataOut, dataOut.Length, serverEndPoint);
                    this.prevMode = mode;
                    this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED;
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED");

                }
                else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_DUNGEN_SEED");
                    int tlen = clientSocket.Receive(receiveData);

                    this.totalDataIn += tlen;
                    this.totalMsgsIn++;

                    byte[] dataIn = receiveData;
                    int msgId = (int)Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DUNGEN_SEED;
                    int msgIdIn = (int)dataIn[0];

                    if (msgIdIn == (int)Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                        int modeIdIn = (int)dataIn[1];
                        int len = vals.Length;

                        for (int i = 0; i < len; i++)
                        {
                            if ((int)vals[i] == modeIdIn)
                            {
                                this.mode = vals[i];
                                wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                                break;
                            }
                        }
                        return;
                    }

                    if (msgId == msgIdIn)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_DUNGEN_SEED msg.");
                        string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                        int dunGenSeed = Converter.toInt(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8] }, REVERSE_FROM_BYTES);
                        long iSentOn = Converter.toLong(new byte[] { dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                        long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        this.serverVersion = version;
                        this.dunGenSeed = dunGenSeed;

                        sync1Recv = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        sync1Tmp = (recvOn - iSentOn);
                        sync1Diff = (sync1Recv - sync1Sent);

                        if (sync1Tmp > 0 && sync1Tmp > sync1Diff)
                        {
                            sync1Diff = sync1Tmp - sync1Diff;
                        }
                        else if (sync1Tmp < 0 && sync1Diff > sync1Tmp)
                        {
                            sync1Diff = sync1Diff - sync1Tmp;
                        }
                        else
                        {
                            sync1Diff = 0;
                        }
                    
                        wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                        wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                        wr(this.clientId + ": " + ip + ":" + port + ": DunGenSeed: " + dunGenSeed);
                        wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": Sync1Diff: " + sync1Diff);

                        if (this.commH == null)
                        {
                            this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                            wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                        }
                        else
                        {
                            this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                        }
                        errorCount = 0;
                    }
                    else
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                        errorCount++;
                        if (errorCount < MAX_ERRORS)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DUNGEN_SEED");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DUNGEN_SEED);
                            }
                        }
                        else
                        {
                            this.Stop();
                        }
                    }

                }
                else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID)
                {
                    rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                    wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_SPAWN_ID msg.");
                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_SPAWN_ID, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + sentOn.Length;
                    sync2Sent = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length), sentOn.Length);

                    NetworkWriteCallBack(dataOut, dataOut.Length, serverEndPoint);
                    this.prevMode = mode;
                    this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_SPAWN_ID;
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUESTED_SPAWN_ID");

                }
                else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_SPAWN_ID)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_SPAWN_ID");
                    int tlen = clientSocket.Receive(receiveData);

                    this.totalDataIn += tlen;
                    this.totalMsgsIn++;

                    byte[] dataIn = receiveData;
                    int msgId = (int)Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_SPAWN_ID;
                    int msgIdIn = (int)dataIn[0];

                    if (msgIdIn == (int)Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                        int modeIdIn = (int)dataIn[1];
                        int len = vals.Length;

                        for (int i = 0; i < len; i++)
                        {
                            if ((int)vals[i] == modeIdIn)
                            {
                                this.mode = vals[i];
                                wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                                break;
                            }
                        }
                        return;
                    }

                    if (msgId == msgIdIn)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_SPAWN_ID msg.");
                        string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                        int spawnId = Converter.toInt(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8] }, REVERSE_FROM_BYTES);
                        long iSentOn = Converter.toLong(new byte[] { dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                        long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        this.serverVersion = version;
                        this.spawnId = spawnId;

                        if (Eds3Client.MOCK_MODE_ON)
                        {
                            mgd = new MockGenDungeon(this.dunGenSeed, this.spawnId);
                            MockGenDungeon.SLEEP = true;
                            mgd.GenerateObjects();
                            mgd.GenerateSpawnObjects();
                        }

                        sync2Recv = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        sync2Tmp = (recvOn - iSentOn);
                        sync2Diff = (sync2Recv - sync2Sent);

                        if (sync2Tmp > 0 && sync2Tmp > sync2Diff)
                        {
                            sync2Diff = sync2Tmp - sync2Diff;
                        }
                        else if (sync2Tmp < 0 && sync2Diff > sync2Tmp)
                        {
                            sync2Diff = sync2Diff - sync2Tmp;
                        }
                        else
                        {
                            sync2Diff = 0;
                        }
                        syncFinalDiff = (sync1Diff + sync2Diff) / 2;
                    
                        if(syncFinalDiff < 0)
                        {
                            syncFinalDiff *= -1;
                        }

                        if(syncFinalDiff > 100)
                        {
                            syncFinalDiff = 15;
                        }

                        wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                        wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                        wr(this.clientId + ": " + ip + ":" + port + ": SpawnId/PlayerId: " + spawnId);
                        wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": Sync1Diff: " + sync1Diff);
                        wr(this.clientId + ": " + ip + ":" + port + ": Sync2Diff: " + sync2Diff);
                        wr(this.clientId + ": " + ip + ":" + port + ": SyncFinalDiff: " + syncFinalDiff);

                        if (this.commH == null)
                        {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            }
                            else
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                            }
                        }
                        else
                        {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2)
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            }
                        }
                        errorCount = 0;
                    }
                    else
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                        errorCount++;
                        if (errorCount < MAX_ERRORS)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_SPAWN_ID");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_SPAWN_ID);
                            }
                        }
                        else
                        {
                            this.Stop();
                        }
                    }

                }
                else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)
                {
                    rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                    wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_SPAWN_ID msg.");
                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DROP_CLIENT, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() - syncFinalDiff, REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + sentOn.Length;

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length), sentOn.Length);

                    NetworkWriteCallBack(dataOut, dataOut.Length, serverEndPoint);
                    this.prevMode = mode;
                    this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DROP_CLIENT;
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");

                    if (!WAIT_FOR_DROP_CLIENT_RESPONSE)
                    {
                        if (this.commH == null)
                        {
                            this.Stop();
                        }
                        else
                        {
                            this.Stop();
                            this.commH.HandleSafeStopComplete();
                        }
                    }

                }
                else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_DROP_CLIENT)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_DROP_CLIENT");
                    int tlen = clientSocket.Receive(receiveData);

                    this.totalDataIn += tlen;
                    this.totalMsgsIn++;

                    byte[] dataIn = receiveData;
                    int msgId = (int)Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DROP_CLIENT;
                    int msgIdIn = (int)dataIn[0];

                    if (msgIdIn == (int)Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                        int modeIdIn = (int)dataIn[1];
                        int len = vals.Length;

                        for (int i = 0; i < len; i++)
                        {
                            if ((int)vals[i] == modeIdIn)
                            {
                                this.mode = vals[i];
                                wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                                break;
                            }
                        }
                        return;
                    }

                    if (msgId == msgIdIn)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_DROP_CLIENT msg.");
                        string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                        int resId = Converter.toInt(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8] }, REVERSE_FROM_BYTES);
                        long iSentOn = Converter.toLong(new byte[] { dataIn[9], dataIn[10], dataIn[11], dataIn[12], dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                        long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();

                        this.serverVersion = version;

                        wr(this.clientId + ": " + ip + ":" + port + ": ServerVersion: " + version);
                        wr(this.clientId + ": " + ip + ":" + port + ": ClientVersion: " + this.version);
                        wr(this.clientId + ": " + ip + ":" + port + ": SpawnId/PlayerId: " + this.spawnId);
                        wr(this.clientId + ": " + ip + ":" + port + ": ResultId: " + resId);
                        wr(this.clientId + ": " + ip + ":" + port + ": SentOn: " + iSentOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": ReceivedOn: " + recvOn);
                        wr(this.clientId + ": " + ip + ":" + port + ": SyncFinalDiff: " + syncFinalDiff);

                        if (this.commH == null)
                        {
                            this.Stop();
                        }
                        else
                        {
                            this.Stop();
                            this.commH.HandleSafeStopComplete();
                        }

                        errorCount = 0;
                    }
                    else
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                        errorCount++;
                        if (errorCount < MAX_ERRORS)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                            }
                        }
                        else
                        {
                            this.Stop();
                        }
                    }
                }
                else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1)
                {
                    rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                    wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1 msg.");
                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() - syncFinalDiff, REVERSE_TO_BYTES);
                    byte[] pId = Converter.toByte(this.spawnId, REVERSE_TO_BYTES);
                    byte[] objs = null;
                    if (this.gomH != null)
                    {
                        objs = this.gomH.GetGameObjectManager().SerializeNetworkGameObjects();
                    }
                    else
                    {
                        objs = this.mgd.SerializeNetworkGameObjects();
                    }
                    byte[] objsLen = Converter.toByte(objs.Length, REVERSE_TO_BYTES);

                    if (objs.Length == 0)
                    {
                        wr("Receiving NO objects to serialize... clearing published list.");
                        //this.SetMode(Eds3ClientMode.WAITING);
                        //return;
                    }

                    wr(this.clientId + ": " + ip + ":" + port + ": Found objsLen: " + objs.Length);
                    int totalLen = msgId.Length + msgVer.Length + sentOn.Length + pId.Length + objsLen.Length + objs.Length;
                    if (totalLen > Eds3Client.DEFAULT_OUT_BUFFER)
                    {
                        int diff = (totalLen - Eds3Client.DEFAULT_OUT_BUFFER);
                        int diffCnt = (int)(diff / (float)this.gowStatic.GetTotalLength());
                        wr(this.clientId + ": " + ip + ":" + port + ": Dropping " + diffCnt + " objects due to serialized object length being too long.");
                        byte[] objs2 = new byte[objs.Length - (diffCnt * this.gowStatic.GetTotalLength())];
                        System.Array.Copy(objs, 0, objs2, 0, objs2.Length);
                        objs = objs2;
                        objsLen = Converter.toByte(objs.Length, REVERSE_TO_BYTES);
                        totalLen = msgId.Length + msgVer.Length + sentOn.Length + pId.Length + objsLen.Length + objs.Length;
                    }

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length), sentOn.Length);
                    System.Array.Copy(pId, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length), pId.Length);
                    System.Array.Copy(objsLen, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length), objsLen.Length);
                    System.Array.Copy(objs, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length + objsLen.Length), objs.Length);

                    NetworkWriteCallBack(dataOut, dataOut.Length, serverEndPoint);
                    this.prevMode = mode;
                    this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1;
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1");

                }
                else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L1");
                    int tlen = clientSocket.Receive(receiveData);

                    this.totalDataIn += tlen;
                    this.totalMsgsIn++;

                    byte[] dataIn = receiveData;
                    int msgId = (int)Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1;
                    int msgIdIn = (int)dataIn[0];

                    if (msgIdIn == (int)Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                        int modeIdIn = (int)dataIn[1];
                        int len = vals.Length;

                        for (int i = 0; i < len; i++)
                        {
                            if ((int)vals[i] == modeIdIn)
                            {
                                this.mode = vals[i];
                                wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                                break;
                            }
                        }
                        return;
                    }

                    if (msgId == msgIdIn)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1 msg.");
                        string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                        long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                        long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                        int spawnId = Converter.toInt(new byte[] { dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                        int objsLen = Converter.toInt(new byte[] { dataIn[17], dataIn[18], dataIn[19], dataIn[20] }, REVERSE_FROM_BYTES);

                        if (objsLen == 0)
                        {
                            wr("Receiving NO objects to deserialize... exiting mode.");
                            this.SetMode(Eds3ClientMode.WAITING);
                            return;
                        }

                        wr(this.clientId + ": ServerVersion: " + version);
                        wr(this.clientId + ": ClientVersion: " + this.version);
                        wr(this.clientId + ": SpawnId/PlayerId: " + this.spawnId);
                        wr(this.clientId + ": SentOn: " + iSentOn);
                        wr(this.clientId + ": ReceivedOn: " + recvOn);
                        wr(this.clientId + ": In SpawnId/PlayerId: " + spawnId);
                        wr(this.clientId + ": Serialized Objects Length: " + objsLen);
                        wr(this.clientId + ": SyncFinalDiff: " + syncFinalDiff);

                        int ttlen = objsLen / this.gowStatic.GetTotalLength();
                        int start = 21;
                        GameObjectWrapper gow;
                        byte[] dtmp;

                        for (int i = 0; i < ttlen; i++)
                        {
                            gow = this.GetEmptyGameObjectWrapper();
                            dtmp = new byte[this.gowStatic.GetTotalLength()];
                            System.Array.Copy(dataIn, start, dtmp, 0, this.gowStatic.GetTotalLength());
                            start += this.gowStatic.GetTotalLength();
                            gow.Deserialize(dtmp);

                            if (this.gomH != null)
                            {
                                this.gomH.GetGameObjectManager().AddInNetworkGameObject(gow);
                            }
                            else
                            {
                                this.mgd.GetGameObjectManager().AddInNetworkGameObject(gow);
                            }

                            if (this.commH != null)
                            {
                                this.commH.HandleDeserializeGameObjectWrapper(objsLen, spawnId, dataIn, (start - gowStatic.GetTotalLength()), ttlen, this, gow);
                            }
                        }

                        wr(this.clientId + ": Deserialized " + ttlen + " game objects.");
                        if (!SERVER_DRIVEN_MODE_ON)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            }
                        }
                        errorCount = 0;
                    }
                    else
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                        errorCount++;
                        if (errorCount < MAX_ERRORS)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                            }
                        }
                        else
                        {
                            this.Stop();
                        }
                    }

                }
                else if (lmode == Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2)
                {
                    rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                    wr(this.clientId + ": " + ip + ":" + port + ": Sending SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2 msg.");
                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() - syncFinalDiff, REVERSE_TO_BYTES);
                    byte[] pId = Converter.toByte(this.spawnId, REVERSE_TO_BYTES);
                    byte[] objs = null;
                    if (this.gomH != null)
                    {
                        objs = this.gomH.GetGameObjectManager().SerializeNetworkGameObjects();
                    }
                    else
                    {
                        objs = this.mgd.SerializeNetworkGameObjects();
                    }
                    byte[] objsLen = Converter.toByte(objs.Length, REVERSE_TO_BYTES);

                    if (objs.Length == 0)
                    {
                        wr("Receiving NO objects to serialize... clearing published list.");
                        //this.SetMode(Eds3ClientMode.WAITING);
                        //return;
                    }

                    long x = 0;
                    long y = 0;
                    long z = 0;
                    GameObjectWrapper lgow = this.GetPlayerChar();

                    if (lgow != null)
                    {
                        x = (long)lgow.GetX();
                        y = (long)lgow.GetY();
                        z = (long)lgow.GetZ();
                    }

                    byte[] bX = Converter.toByte(x, REVERSE_TO_BYTES);
                    byte[] bY = Converter.toByte(y, REVERSE_TO_BYTES);
                    byte[] bZ = Converter.toByte(z, REVERSE_TO_BYTES);

                    wr(this.clientId + ": " + ip + ":" + port + ": Found objsLen: " + objs.Length);
                    int totalLen = msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length + bY.Length + bZ.Length + objsLen.Length + objs.Length;
                    if (totalLen > Eds3Client.DEFAULT_OUT_BUFFER)
                    {
                        int diff = (totalLen - Eds3Client.DEFAULT_OUT_BUFFER);
                        int diffCnt = (int)(diff / (float)this.gowStatic.GetTotalLength());
                        wr(this.clientId + ": " + ip + ":" + port + ": Dropping " + diffCnt + " objects due to serialized object length being too long.");
                        byte[] objs2 = new byte[objs.Length - (diffCnt * this.gowStatic.GetTotalLength())];
                        System.Array.Copy(objs, 0, objs2, 0, objs2.Length);
                        objs = objs2;
                        objsLen = Converter.toByte(objs.Length);
                        totalLen = msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length + bY.Length + bZ.Length + objsLen.Length + objs.Length;
                    }

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length), sentOn.Length);
                    System.Array.Copy(pId, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length), pId.Length);
                    System.Array.Copy(bX, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length), bX.Length);
                    System.Array.Copy(bY, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length), bY.Length);
                    System.Array.Copy(bZ, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length + bY.Length), bZ.Length);
                    System.Array.Copy(objsLen, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length + bY.Length + bZ.Length), objsLen.Length);
                    System.Array.Copy(objs, 0, dataOut, (msgId.Length + msgVer.Length + sentOn.Length + pId.Length + bX.Length + bY.Length + bZ.Length + objsLen.Length), objs.Length);

                    NetworkWriteCallBack(dataOut, dataOut.Length, serverEndPoint);
                    this.prevMode = mode;
                    this.mode = Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2;
                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L2");
                }
                else if (lmode == Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2");
                    int tlen = clientSocket.Receive(receiveData);

                    this.totalDataIn += tlen;
                    this.totalMsgsIn++;

                    byte[] dataIn = receiveData;
                    int msgId = (int)Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2;
                    int msgIdIn = (int)dataIn[0];

                    if (msgIdIn == (int)Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_SET_MODE_CMD");
                        int modeIdIn = (int)dataIn[1];
                        int len = vals.Length;

                        for (int i = 0; i < len; i++)
                        {
                            if ((int)vals[i] == modeIdIn)
                            {
                                this.mode = vals[i];
                                wr(this.clientId + ": " + ip + ":" + port + ": Setting Mode: " + this.mode);
                                break;
                            }
                        }
                        return;
                    }

                    string version = "";
                    long iSentOn = 0;
                    long recvOn = 0;
                    int spawnId = 0;
                    byte[] bIp = null;
                    string iIp = "";
                    int objsLen = 0;
                    int start = 0;
                    int cnt = 1;
                    int ttlen = 0;
                    int lstart = 0;
                    GameObjectWrapper gow;
                    byte[] dtmp;

                    if (msgId == msgIdIn)
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Receiving SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg.");

                        msgIdIn = (int)dataIn[start];
                        start += 1;

                        while (msgId == msgIdIn)
                        {
                            version = Converter.toString(new byte[] { dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3] }, REVERSE_FROM_BYTES);
                            start += 4;

                            iSentOn = Converter.toLong(new byte[] { dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3], dataIn[start + 4], dataIn[start + 5], dataIn[start + 6], dataIn[start + 7] }, REVERSE_FROM_BYTES);
                            start += 8;

                            recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            spawnId = Converter.toInt(new byte[] { dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3] }, REVERSE_FROM_BYTES);
                            start += 4;

                            bIp = new byte[Eds3Server.IP_SER_LENGTH];
                            System.Array.Copy(dataIn, start, bIp, 0, bIp.Length);
                            iIp = Converter.toString(bIp);
                            start += bIp.Length;

                            objsLen = Converter.toInt(new byte[] { dataIn[start], dataIn[start + 1], dataIn[start + 2], dataIn[start + 3] }, REVERSE_FROM_BYTES);
                            start += 4;

                            if (objsLen == 0)
                            {
                                wr("Receiving NO objects to deserialize... exiting mode.");
                                this.SetMode(Eds3ClientMode.WAITING);
                                return;
                            }

                            wr(this.clientId + ": ServerVersion: " + version);
                            wr(this.clientId + ": ClientVersion: " + this.version);
                            wr(this.clientId + ": SpawnId/PlayerId: " + this.spawnId);
                            wr(this.clientId + ": SentOn: " + iSentOn);
                            wr(this.clientId + ": ReceivedOn: " + recvOn);
                            wr(this.clientId + ": In SpawnId/PlayerId: " + spawnId);
                            wr(this.clientId + ": In IP Address: " + iIp);
                            wr(this.clientId + ": Serialized Objects Length: " + objsLen);
                            wr(this.clientId + ": Frame: " + cnt);
                            wr(this.clientId + ": LStart: " + start);
                            wr(this.clientId + ": SyncFinalDiff: " + syncFinalDiff);

                            ttlen = objsLen / this.gowStatic.GetTotalLength();
                            lstart = start;

                            for (int i = 0; i < ttlen; i++)
                            {
                                gow = this.GetEmptyGameObjectWrapper();
                                dtmp = new byte[this.gowStatic.GetTotalLength()];

                                System.Array.Copy(dataIn, lstart, dtmp, 0, this.gowStatic.GetTotalLength());
                                lstart += this.gowStatic.GetTotalLength();
                                gow.Deserialize(dtmp);

                                if (this.gomH != null)
                                {
                                    this.gomH.GetGameObjectManager().AddInNetworkGameObject(gow);
                                }
                                else
                                {
                                    this.mgd.GetGameObjectManager().AddInNetworkGameObject(gow);
                                }

                                if (this.commH != null)
                                {
                                    this.commH.HandleDeserializeGameObjectWrapper(objsLen, spawnId, dataIn, lstart, ttlen, this, gow);
                                }
                            }

                            wr(this.clientId + ": Deserialized " + ttlen + " game objects.");

                            start += objsLen;
                            msgIdIn = (int)dataIn[start];
                            start += 1;
                            cnt += 1;

                            if (start >= dataIn.Length)
                            {
                                break;
                            }
                        }

                        if (!SERVER_DRIVEN_MODE_ON)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_RECEIVE_REQUESTED_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": ExitMode: CLIENT_RECEIVE_REQUEST_OBJ_PUBLISH_L2");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            }
                        }
                        errorCount = 0;
                    }
                    else
                    {
                        wr(this.clientId + ": " + ip + ":" + port + ": Unexpected msg id for this mode. Exiting back to start mode for this msg chain.");
                        errorCount++;
                        if (errorCount < MAX_ERRORS)
                        {
                            if (this.commH == null)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(true, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                            }
                        }
                        else
                        {
                            this.Stop();
                        }
                    }

                }
                else if (lmode == Eds3ClientMode.WAITING)
                {
                    //rootMode = lmode;
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: WAITING");
                    wr(this.clientId + ": " + ip + ":" + port + ": Receiving WAITING msg.");

                    if (Eds3Client.MOCK_MODE_ON)
                    {
                        Thread.Sleep(10);
                    }

                    if (this.commH == null)
                    {
                        this.SetMode(Eds3ClientMode.RESTING);
                        wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: RESTING");
                    }
                    else
                    {
                        this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.RESTING);
                    }
                }
                else if (lmode == Eds3ClientMode.RESTING)
                {
                    wr(this.clientId + ": " + ip + ":" + port + ": EnterMode: RESTING");
                    if (MOCK_MODE_ON)
                    {
                        Thread.Sleep(1000);
                        if (this.mgd != null)
                        {
                            this.mgd.RandomlyAdjustObjects();
                        }
                    }

                    if (this.commH == null)
                    {
                        if (MOCK_MODE_ON)
                        {
                            wr("Found total messages of " + this.totalMsgsOut + " out of " + Eds3Client.MAX_MOCK_MESSAGES_OUT);
                            if (this.totalMsgsOut >= Eds3Client.MAX_MOCK_MESSAGES_OUT)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                            }
                            else
                            {
                                if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2)
                                {
                                    this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                                }
                                else
                                {
                                    this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1");
                                }
                            }
                        }
                        else
                        {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L1)
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_l1");
                            }
                            else
                            {
                                this.SetMode(Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            }
                        }
                    }
                    else
                    {
                        if (MOCK_MODE_ON)
                        {
                            wr("Found total messages of " + this.totalMsgsOut + " out of " + Eds3Client.MAX_MOCK_MESSAGES_OUT);
                            if (this.totalMsgsOut >= Eds3Client.MAX_MOCK_MESSAGES_OUT)
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_DROP_CLIENT");
                            }
                            else
                            {
                                if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L2)
                                {
                                    this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                                }
                                else
                                {
                                    this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                    wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH");
                                }
                            }
                        }
                        else
                        {
                            if (Eds3Client.OBJ_SER_MODE == Eds3ClientObjSerMode.L1)
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L1);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH");
                            }
                            else
                            {
                                this.commH.HandleCurrentModeResult(false, this, this.mode, this.prevMode, Eds3ClientMode.CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2);
                                wr(this.clientId + ": " + ip + ":" + port + ": ExitMode: CLIENT_SEND_REQUEST_OBJ_PUBLISH_L2");
                            }
                        }
                    }
                }
                else
                {
                    if (Eds3Client.MOCK_MODE_ON)
                    {
                        Thread.Sleep(10);
                    }

                    if (this.commH != null)
                    {
                        this.commH.HandleProcessModeUnknown(receiveData, clientSocket, this);
                    }
                }

                this.globalErrorCount = 0;
            }
            catch (Exception e)
            {
                if (!this.shuttingDown)
                {
                    //wr("Globel exception encountered " + this.globalErrorCount + " found out of " + Eds3Client.MAX_GLOBAL_ERRORS);
                    if (this.commH != null)
                    {
                        this.commH.HandleNetworkError(e);
                    }

                    this.globalErrorCount++;
                    if (this.globalErrorCount > Eds3Client.MAX_GLOBAL_ERRORS)
                    {
                        this.Stop();
                    }

                    this.forceNextMode = true;
                    this.nextMode = rootMode;
                    SetMode(rootMode);

                    if (e != null)
                    {
                        if (MOCK_MODE_ON && e.Message.Contains("Operation timed out") == false)
                        {
                            wr(e.Message);
                            PrintStackTrace(e.StackTrace);
                        }
                    }
                }
            }
        }

        public void SafeStop()
        {
            this.startShuttingDown = true;
            this.SetForceNextMode(Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT);
        }

        private void AsyncNetworkReadCallBack(byte[] dataIn, int tlen, EndPoint ServerEndPoint)
        {
            try
            {
                lock (this)
                {
                    this.ancDataIn = dataIn;
                    this.ancLen = tlen;
                    this.ancServerEndPoint = ServerEndPoint;
                }
                ThreadStart ts = new ThreadStart(NetworkReadCallBack);
                Thread trd = new Thread(ts);
                trd.Start();
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
                throw e;
            }
        }

        private void AsyncNetworkWriteCallBack(byte[] dataOut, int tlen, EndPoint ServerEndPoint)
        {
            try
            {
                lock (this)
                {
                    this.ancDataOut = dataOut;
                    this.ancLen = tlen;
                    this.ancServerEndPoint = ServerEndPoint;
                }
                ThreadStart ts = new ThreadStart(NetworkWriteCallBack);
                Thread trd = new Thread(ts);
                trd.Start();
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
                throw e;
            }
        }

        public void NetworkReadCallBack(byte[] dataIn, int tlen, EndPoint ServerEndPoint)
        {
            lock (this)
            {
                this.ancDataIn = dataIn;
                this.ancLen = tlen;
                this.ancServerEndPoint = ServerEndPoint;
            }
            this.NetworkReadCallBack();
        }

        public void NetworkReadCallBack()
        {
            //Retrieve the state object and the handler socket.
            wr("");
            wr("");
            wr("NetworkReadCallBack");

            try
            {
                byte[] dataIn = null;
                int tlen = -1;
                EndPoint ServerEndPoint = null;

                lock (this)
                {
                    dataIn = this.ancDataIn;
                    tlen = this.ancLen;
                    ServerEndPoint = this.ancServerEndPoint;
                }

                if (this.shuttingDown)
                {
                    return;
                }

                if (this.paused)
                {
                    return;
                }
            }
            catch (Exception e)
            {
                wr("Exception caught in NetworkReadCallBack.");
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
                throw e;
            }
        }

        public void NetworkWriteCallBack(byte[] dataOut, int tlen, EndPoint ServerEndPoint)
        {
            lock (this)
            {
                this.ancDataOut = dataOut;
                this.ancLen = tlen;
                this.ancServerEndPoint = ServerEndPoint;
            }
            this.NetworkWriteCallBack();
        }

        public void NetworkWriteCallBack()
        {
            //Retrieve the state object and the handler socket.
            wr("");
            wr("");
            wr("NetworkWriteCallBack");

            try
            {
                byte[] dataOut = null;
                int tlen = -1;
                EndPoint ServerEndPoint = null;

                lock (this)
                {
                    dataOut = this.ancDataOut;
                    tlen = this.ancLen;
                    ServerEndPoint = this.ancServerEndPoint;
                }

                if (this.shuttingDown)
                {
                    return;
                }

                if (this.paused)
                {
                    return;
                }

                wr("DataOutLen: " + dataOut.Length + " TLen: " + tlen);
                this.totalDataOut += dataOut.Length;
                this.totalMsgsOut++;
                this.clientSocket.SendTo(dataOut, ServerEndPoint);
            }
            catch (Exception e)
            {
                wr("Exception caught in NetworkWriteCallBack.");
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
                throw e;
            }
        }
        #endregion
    }
}