﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Threading;
using System.Net.Http;
using System.Threading.Tasks;
using com.middlemind.Eds3.unity;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3
{
    public class Eds3Server
    {
        #region static_vars
        //STATIC VARIABLES
        public const int DUN_GEN_SEED_ERROR = int.MaxValue;
        public const int SPAWN_ID_ERROR = int.MaxValue;
        public const string TEST_URL = "http://www.google.com";

        public const int MIB_INT = 1048576;
        public const int KIB_INT = 1024;
        public const int CHECK_LOG_SIZE_TICKS = 25;

        public const int CHECK_LOG_SIZE = 5000000;
        public const int DEFAULT_MAX_GAME_ROOMS = 10;
        public const int DEFAULT_MAX_GAME_ROOM_PLAYERS = 10;

        public const bool DEFAULT_DEBUG_ON = false;
        public const int DEFAULT_PORT = 49986;
        public const int DEFAULT_TIMEOUT = 15000;

        public const int DEFAULT_CONNECTION_TIMEOUT = 15000;
        public const string DEFAULT_TITLE = "Eds3Server";

        public static string DEFAULT_LOCAL_IP_ADDRESS = GetLocalIPAddressStat();
        public const int DEFAULT_IN_BUFFER = 9000;
        public const int DEFAULT_OUT_BUFFER = 9000;

        private const int MIN_PORT = 0;
        private const int MAX_PORT = (short.MaxValue * 2);

        private const int MIN_TIMEOUT = 0;
        private const int MAX_TIMEOUT = (short.MaxValue * 2);

        private const int MAX_IN_BUFFER_LENGTH = 9000;
        private const int MIN_IN_BUFFER_LENGTH = (KIB_INT * 1);

        private const int MAX_OUT_BUFFER_LENGTH = 9000;
        private const int MIN_OUT_BUFFER_LENGTH = (KIB_INT * 1);

        private const bool FORCE_PORT_USE = true;
        public static bool ASYNC_ON = true;

        public static bool OBJ_PUBLISH_L2_DIST_LIMIT_ON = true;
        public static double OBJ_PUBLISH_L2_DIST_MAX_SQR = (1000.0 * 1000.0);
        public static double OBJ_PUBLISH_L2_DIST_MAX = 1000.0;

        public static int IP_SER_LENGTH = 32;
        public static bool FORCE_FLUSH = false;
        public static bool SERVER_DRIVEN_MODE_ON = false;

        private static Random rand = new System.Random((int)System.DateTime.Now.Ticks);
        private static bool REVERSE_TO_BYTES = false;
        private static bool REVERSE_FROM_BYTES = false;
        #endregion

        #region local_vars
        //LOCAL VARIABLES
        private int totalDataIn = 0;
        private int totalMsgsIn = 0;
        private int totalDataOut = 0;
        private int totalMsgsOut = 0;

        private int inBufferSize = DEFAULT_IN_BUFFER;
        private int outBufferSize = DEFAULT_OUT_BUFFER;
        private string version = "0101";

        private string appName = DEFAULT_TITLE;
        private string title = DEFAULT_TITLE;
        private int port = DEFAULT_PORT;

        private int timeout = DEFAULT_TIMEOUT;

        private Dictionary<string, GameRoom> players = null;
        private List<GameRoom> gameRooms = null;
        private bool shuttingDown = false;

        private int maxGameRooms = DEFAULT_MAX_GAME_ROOMS;
        private int maxGameRoomPlayers = DEFAULT_MAX_GAME_ROOM_PLAYERS;
        private string rootDir = ".";
        private string appDir = "";

        private string configFileName = "config.txt";
        private string configFilePath = "";
        private string debugFileName = "debug.txt";

        private string debugFilePath = "";
        private bool debugOn = DEFAULT_DEBUG_ON;
        private StreamWriter debug = null;

        private StreamWriter fw = null;
        private int wrCount = 0;
        private string localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
        private IPEndPoint endPoint;
        private EndPoint serverEndPoint;
        private Socket serverSocket;

        private byte[] ancDataIn;
        private byte[] ancDataOut;
        private int ancLen;
        private EndPoint ancClientEndPoint;
        private bool closeAllClients = false;
        private HandleEds3ServerComm commH = null;
        private GameObjectWrapper gowStatic = null;
        #endregion

        #region enumerations
        public enum Eds3ServerMode
        {
            SERVER_RECEIVE_REQUEST_DUNGEN_SEED,
            SERVER_SEND_REQUESTED_DUNGEN_SEED,
            SERVER_RECEIVE_REQUEST_SPAWN_ID,
            SERVER_SEND_REQUESTED_SPAWN_ID,
            SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1,
            SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1,
            SERVER_SEND_SET_MODE_CMD,
            SERVER_RECEIVE_REQUEST_DROP_CLIENT,
            SERVER_SEND_REQUESTED_DROP_CLIENT,
            SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2,
            SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2
        }

        //Helper used for looping over the mode enumeration.
        private Eds3ServerMode[] vals = new Eds3Server.Eds3ServerMode[]
        {
            Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DUNGEN_SEED,
            Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DUNGEN_SEED,
            Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_SPAWN_ID,
            Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_SPAWN_ID,
            Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1,
            Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1,
            Eds3Server.Eds3ServerMode.SERVER_SEND_SET_MODE_CMD,
            Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_DROP_CLIENT,
            Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_DROP_CLIENT,
            Eds3Server.Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2,
            Eds3Server.Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2
        };
        #endregion

        public Eds3Server()
        {
            appDir = rootDir + System.IO.Path.DirectorySeparatorChar + appName;
            configFilePath = appDir + System.IO.Path.DirectorySeparatorChar + configFileName;
            debugFilePath = appDir + System.IO.Path.DirectorySeparatorChar + debugFileName;

            System.Console.Out.WriteLine("");
            System.Console.Out.WriteLine("App Dir: " + appDir);
            System.Console.Out.WriteLine("Config Path: " + configFilePath);
            System.Console.Out.WriteLine("Debug Path: " + debugFilePath);
            System.Console.Out.WriteLine("Version: " + version);

            LoadSettings();
            CreateDebugFile();
            PrintEds3ModeValues();

            players = new Dictionary<string, GameRoom>();
            gameRooms = new List<GameRoom>();
            this.gowStatic = this.GetEmptyGameObjectWrapper();
        }

        public Eds3Server(HandleEds3ServerComm CommH)
        {
            appDir = rootDir + System.IO.Path.DirectorySeparatorChar + appName;
            configFilePath = appDir + System.IO.Path.DirectorySeparatorChar + configFileName;
            debugFilePath = appDir + System.IO.Path.DirectorySeparatorChar + debugFileName;
            this.commH = CommH;

            System.Console.Out.WriteLine("");
            System.Console.Out.WriteLine("App Dir: " + appDir);
            System.Console.Out.WriteLine("Config Path: " + configFilePath);
            System.Console.Out.WriteLine("Debug Path: " + debugFilePath);
            System.Console.Out.WriteLine("Version: " + version);

            LoadSettings();
            CreateDebugFile();
            PrintEds3ModeValues();

            players = new Dictionary<string, GameRoom>();
            gameRooms = new List<GameRoom>();
            this.gowStatic = this.GetEmptyGameObjectWrapper();

            if (this.commH != null)
            {
                this.commH.HandleEds3ServerInit(this);
            }
        }

        #region utility_methods
        public void PrintEds3ModeValues()
        {
            int len = this.vals.Length;

            for (int i = 0; i < len; i++)
            {
                wr("Idx: " + i + " Eds3Msg Value: " + (int)vals[i]);
            }
        }

        //Attempts to read the expected config file and load the settings.
        //Running the service in cmd line mode shows the expected file paths and config file format.
        private void LoadSettings()
        {
            try
            {
                System.Console.Out.WriteLine("");
                System.Console.Out.WriteLine("Loading settings...");
                System.Console.Out.WriteLine("Looking for data directory: " + this.appDir);
                DirectoryInfo dInf = new DirectoryInfo(this.appDir);
                FileInfo fInf;
                if (dInf.Exists == false)
                {
                    System.Console.Out.WriteLine("Creating the app directory, this is where the config file should go, config.txt.");
                    System.IO.Directory.CreateDirectory(this.appDir);
                }

                System.Console.Out.WriteLine("");
                System.Console.Out.WriteLine("Looking for config file: " + this.configFilePath);
                System.Console.Out.WriteLine("Entries in the config file are as follows.");
                System.Console.Out.WriteLine("port=[" + DEFAULT_PORT + ": The desired port number.]");
                System.Console.Out.WriteLine("debug=[" + (DEFAULT_DEBUG_ON ? "1" : "0") + ": Set to 1/true for true, 0/false for false.]");
                System.Console.Out.WriteLine("connection_timeout_ms=[" + DEFAULT_TIMEOUT + ": The number of miliseconds to wait for socket timeout.]");
                System.Console.Out.WriteLine("in_buffer_size=[" + DEFAULT_IN_BUFFER + ": The number of bytes for the input connection buffer.]");
                System.Console.Out.WriteLine("out_buffer_size=[" + DEFAULT_OUT_BUFFER + ": The number of bytes for the output connection buffer.]");
                System.Console.Out.WriteLine("local_ip_address=[" + DEFAULT_LOCAL_IP_ADDRESS + ": The local IP address to bind to.]");

                System.Console.Out.WriteLine("");
                fInf = new FileInfo(this.configFilePath);
                if (fInf.Exists == true)
                {
                    System.Console.Out.WriteLine("Found a config file.");
                    System.Console.Out.WriteLine("Loading config file values...");

                    StreamReader sr = new StreamReader(this.configFilePath);
                    string line = sr.ReadLine();
                    while (line != null)
                    {
                        bool success = false;
                        int v = 0;
                        IPAddress ip;
                        string tmp = "";
                        string[] pair = line.Split('=');
                        if (pair.Length == 2)
                        {
                            string key = pair[0];
                            string val = pair[1];
                            if ("PORT".Equals(key.ToUpper()))
                            {
                                success = false;

                                try
                                {
                                    v = int.Parse(val);
                                    success = true;
                                }
                                catch (Exception e)
                                {

                                }

                                if (success == true)
                                {
                                    if (port >= MIN_PORT && port <= MAX_PORT)
                                    {
                                        port = v;
                                        System.Console.Out.WriteLine("Found port: " + this.port);
                                    }
                                    else
                                    {
                                        port = DEFAULT_PORT;
                                        System.Console.Out.WriteLine("Port outside bounds, using default timeout: " + this.port);
                                    }
                                }
                                else
                                {
                                    port = DEFAULT_PORT;
                                    System.Console.Out.WriteLine("Could not parse port, using default port: " + this.port);
                                }
                            }
                            else if ("DEBUG".Equals(key.ToUpper()))
                            {
                                tmp = val.ToUpper();
                                if ("1".Equals(tmp) || "TRUE".Equals(tmp))
                                {
                                    this.debugOn = true;
                                }
                                else
                                {
                                    this.debugOn = false;
                                }
                                System.Console.Out.WriteLine("Found debug: " + this.debugOn);
                            }
                            else if ("CONNECTION_TIMEOUT_MS".Equals(key.ToUpper()))
                            {
                                success = false;
                                try
                                {
                                    v = int.Parse(val);
                                    success = true;
                                }
                                catch (Exception e)
                                {

                                }

                                if (success == true)
                                {
                                    if (v >= MIN_TIMEOUT && v <= MAX_TIMEOUT)
                                    {
                                        timeout = v;
                                        System.Console.Out.WriteLine("Found timeout: " + this.timeout);
                                    }
                                    else
                                    {
                                        timeout = DEFAULT_TIMEOUT;
                                        System.Console.Out.WriteLine("Timeout outside bounds, using default timeout: " + this.timeout);
                                    }
                                }
                                else
                                {
                                    timeout = DEFAULT_TIMEOUT;
                                    System.Console.Out.WriteLine("Could not parse timeout, using default timeout: " + this.timeout);
                                }
                            }
                            else if ("OUT_BUFFER_SIZE".Equals(key.ToUpper()))
                            {
                                success = false;
                                try
                                {
                                    v = int.Parse(val);
                                    success = true;
                                }
                                catch (Exception e)
                                {

                                }

                                if (success == true)
                                {
                                    if (v >= MIN_OUT_BUFFER_LENGTH && v <= MAX_OUT_BUFFER_LENGTH)
                                    {
                                        outBufferSize = v;
                                        System.Console.Out.WriteLine("Found out buffer size: " + this.outBufferSize);
                                    }
                                    else
                                    {
                                        outBufferSize = DEFAULT_OUT_BUFFER;
                                        System.Console.Out.WriteLine("Out buffer size outside bounds, using default out buffer size: " + this.outBufferSize);
                                    }
                                }
                                else
                                {
                                    outBufferSize = DEFAULT_OUT_BUFFER;
                                    System.Console.Out.WriteLine("Could not parse out buffer size, using default out buffer size: " + this.outBufferSize);
                                }
                            }
                            else if ("IN_BUFFER_SIZE".Equals(key.ToUpper()))
                            {
                                success = false;
                                try
                                {
                                    v = int.Parse(val);
                                    success = true;
                                }
                                catch (Exception e)
                                {

                                }

                                if (success == true)
                                {
                                    if (v >= MIN_IN_BUFFER_LENGTH && v <= MAX_IN_BUFFER_LENGTH)
                                    {
                                        inBufferSize = v;
                                        System.Console.Out.WriteLine("Found in buffer size: " + this.inBufferSize);
                                    }
                                    else
                                    {
                                        inBufferSize = DEFAULT_IN_BUFFER;
                                        System.Console.Out.WriteLine("In buffer size outside bounds, using default in buffer size: " + this.inBufferSize);
                                    }
                                }
                                else
                                {
                                    inBufferSize = DEFAULT_IN_BUFFER;
                                    System.Console.Out.WriteLine("Could not parse in buffer size, using default in buffer size: " + this.inBufferSize);
                                }
                            }
                            else if ("LOCAL_IP_ADDRESS".Equals(key.ToUpper()))
                            {
                                success = false;
                                try
                                {
                                    int.Parse(val.Replace(".", ""));
                                    ip = IPAddress.Parse(val);
                                    success = true;
                                }
                                catch (Exception e)
                                {

                                }

                                if (success == true)
                                {
                                    localIpAddress = val;
                                    System.Console.Out.WriteLine("Found local ip address: " + this.localIpAddress);
                                }
                                else
                                {
                                    localIpAddress = DEFAULT_LOCAL_IP_ADDRESS;
                                    System.Console.Out.WriteLine("Could not parse local ip address, using default local ip address: " + this.localIpAddress);
                                }
                            }
                        }
                        line = sr.ReadLine();
                    }
                }
                else
                {
                    System.Console.Out.WriteLine("Could not find a config file.");
                    System.Console.Out.WriteLine("Loading default values...");
                    port = DEFAULT_PORT;
                    debugOn = DEFAULT_DEBUG_ON;
                    timeout = DEFAULT_TIMEOUT;
                    outBufferSize = DEFAULT_OUT_BUFFER;
                    inBufferSize = DEFAULT_IN_BUFFER;
                }
            }
            catch (Exception e)
            {
                PrintStackTrace(e.StackTrace);
            }
            PrintConfigAsync();
        }

        public async Task<bool> PrintConfigAsync()
        {
            System.Console.Out.WriteLine("==========CURRENT CONFIG===========");
            System.Console.Out.WriteLine("Version: " + version);
            System.Console.Out.WriteLine("Found port: " + this.port);
            System.Console.Out.WriteLine("Found debug: " + this.debugOn);
            System.Console.Out.WriteLine("Found connection_timeout_ms: " + this.timeout + " ms");
            System.Console.Out.WriteLine("Found in_buffer_size: " + this.inBufferSize + " bytes");
            System.Console.Out.WriteLine("Found out_buffer_size: " + this.outBufferSize + " bytes");

            Task<bool> tb = GetIsNetworkAvailableAsync();
            bool b = await tb;

            if (b)
            {
                GetLocalIPAddress();
                return true;
            }
            else
            {
                System.Console.Out.WriteLine("Network connection not found.");
                return false;
            }
        }

        public async Task<bool> GetIsNetworkAvailableAsync()
        {
            try
            {
                Uri url = new Uri(TEST_URL);
                HttpClient client = new HttpClient();
                string responseString = await client.GetStringAsync(url);
                if (responseString != null && responseString.Equals("") != true)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            catch (Exception e)
            {
                PrintStackTrace(e.StackTrace);
            }
            return false;
        }

        public static string GetLocalIPAddressStat()
        {
            try
            {
                var host = Dns.GetHostEntry(Dns.GetHostName());
                foreach (var ip in host.AddressList)
                {
                    if (ip.AddressFamily == AddressFamily.InterNetwork)
                    {
                        return ip.ToString();
                    }
                }
            }
            catch (Exception e)
            {
                PrintStackTrace(e.StackTrace);
            }
            return "0.0.0.0";
        }

        public string GetLocalIPAddress()
        {
            try
            {
                var host = Dns.GetHostEntry(Dns.GetHostName());
                foreach (var ip in host.AddressList)
                {
                    if (ip.AddressFamily == AddressFamily.InterNetwork)
                    {
                        return ip.ToString();
                    }
                }
            }
            catch (Exception e)
            {
                PrintStackTrace(e.StackTrace);
            }
            return "0.0.0.0";
        }

        public void PrintState()
        {
            System.Console.Out.WriteLine("==========CURRENT STATE===========");
            System.Console.Out.WriteLine("GameRooms: " + this.gameRooms.Count);
            System.Console.Out.WriteLine("TotalPlayers: " + this.CountTotalPlayers());
        }

        private int CountTotalPlayers()
        {
            int total = 0;
            int len = gameRooms.Count;
            for (int i = 0; i < len; i++)
            {
                total += gameRooms[i].players.Count;
            }
            return total;
        }

        private void PrintStackTrace(string[] st)
        {
            if (st != null)
            {
                int len = st.Length;
                for (int i = 0; i < len; i++)
                {
                    wr(st[i].ToString());
                }
            }
        }

        public static void PrintStackTrace(string s)
        {
            if (s != null)
            {
                System.Console.Out.WriteLine(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() + ": " + s);
                System.Diagnostics.Debug.WriteLine(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() + ": " + s);
            }
        }

        public void wr(string s)
        {
            if (this.debugOn == true)
            {
                System.Console.Out.WriteLine(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds() + ": " + s);
                if (s == null)
                {
                    s = "";
                }

                if (this.debug != null)
                {
                    try
                    {
                        this.debug.Write(s);
                        this.debug.Write(this.debug.NewLine);
                        if (FORCE_FLUSH)
                        {
                            this.debug.Flush();
                        }
                    }
                    catch (Exception e)
                    {
                        PrintStackTrace(e.StackTrace);
                        this.debugOn = false;
                        this.debug = null;
                    }
                    wrCount++;
                }

                if (wrCount > 0 && wrCount % CHECK_LOG_SIZE_TICKS == 0)
                {
                    FileInfo fInf = new FileInfo(this.debugFilePath);
                    if (fInf.Length >= CHECK_LOG_SIZE)
                    {
                        try
                        {
                            this.debug.Flush();
                            this.debug.Close();
                        }
                        catch (Exception e)
                        {
                            PrintStackTrace(e.StackTrace);
                            this.debugOn = false;
                            this.debug = null;
                        }
                        CreateDebugFile(false);
                    }
                }
            }
        }
        #endregion

        #region log_file_methods
        //Opens up a new logging file.
        private void CreateDebugFile()
        {
            CreateDebugFile(true);
        }

        //Opens up a new logging file with the append setting.
        private void CreateDebugFile(bool append)
        {
            if (this.debugOn == true)
            {
                try
                {
                    FileInfo f = new FileInfo(this.debugFilePath);
                    if (f.Exists == false)
                    {
                        f.Create();
                        this.debug = new StreamWriter(this.debugFilePath, false);
                    }
                    else
                    {

                        this.debug = new StreamWriter(this.debugFilePath, append);
                    }

                }
                catch (Exception e)
                {
                    PrintStackTrace(e.StackTrace);
                    this.debugOn = false;
                    this.debug = null;
                }
                this.wrCount = 0;
            }
        }

        private void CloseDebugFile()
        {
            try
            {
                this.debug.Flush();
            }
            catch (Exception e)
            {
            }

            try
            {
                this.debug.Close();
            }
            catch (Exception e)
            {
            }

            try
            {
                this.fw.Flush();
            }
            catch (Exception e)
            {
            }

            try
            {
                this.fw.Close();
            }
            catch (Exception e)
            {
            }
        }
        #endregion

        #region getter_and_setter_methods
        public GameObjectWrapper GetEmptyGameObjectWrapper()
        {
            if (this.commH != null)
            {
                return this.commH.GetEmptyGameObjectWrapper();
            }
            else
            {
                return new GameObjectWrapperDefault();
            }
        }

        public HandleEds3ServerComm GetCommH()
        {
            return commH;
        }

        public void SetCommH(HandleEds3ServerComm commH)
        {
            this.commH = commH;
        }

        public void SetPort(int p)
        {
            this.port = p;
        }

        public void SetPort(String p)
        {
            int tmp = -1;
            bool b = int.TryParse(p, out tmp);
            if (b)
            {
                this.port = tmp;
            }
        }

        public bool IsShuttingDown()
        {
            return this.shuttingDown;
        }

        //Returns the title of the service that has been set.
        public string GetTitle()
        {
            return title;
        }

        public Dictionary<string, GameRoom> GetPlayers()
        {
            return this.players;
        }

        //Returns true if logging is turned on.
        public bool GetDebugOn()
        {
            return debugOn;
        }

        //Returns the full path to the logging file.
        public string GetDebugFilePath()
        {
            return debugFilePath;
        }

        //Returns the file name of the logging file.
        public string GetDebugFileName()
        {
            return debugFileName;
        }

        //Returns the application name of this service.
        public string GetAppName()
        {
            return appName;
        }

        //Returns the port number this service is using.
        public int GetPort()
        {
            return port;
        }

        //Returns the full path of the root directory the service is using.
        public string GetRootDir()
        {
            return rootDir;
        }

        //Returns the full path of the service config/log directory.
        public string GetAppDir()
        {
            return appDir;
        }

        //Returns the name of the config file this application is using.
        public string GetConfigFileName()
        {
            return configFileName;
        }

        //Returns the full path of the expected config file.
        public string GetConfigFilePath()
        {
            return configFilePath;
        }

        //Returns true if this service is shutting down.
        public bool GetShuttingDown()
        {
            return shuttingDown;
        }

        public string GetLocalIpAddress()
        {
            return this.localIpAddress;
        }

        public void SetLocalIpAddress(string ip)
        {
            this.localIpAddress = ip;
        }
        #endregion

        #region network_helper_methods
        private int GetNewDunGenSeed()
        {
            return (int)(rand.Next(Eds3Server.DUN_GEN_SEED_ERROR) - 1);
        }

        private bool AssignGameRoom(Player p)
        {
            if (p == null)
            {
                return false;
            }

            int len = this.gameRooms.Count;
            GameRoom r = null;
            bool found = false;
            int id = 0;
            string ip = p.ip;
            int port = p.port;

            for (int i = 0; i < len; i++)
            {
                r = this.gameRooms[i];
                if (r != null && r.players.ContainsKey(p.ip))
                {
                    wr(ip + ":" + port + ": Player already exists.");
                    return true;
                }

                if (r != null && r.players.Count < this.maxGameRoomPlayers && !found)
                {
                    wr(ip + ":" + port + ": Player doesn't already exist, but room does.");
                    id = i;
                    found = true;
                }
            }

            if (!found)
            {
                if (this.gameRooms.Count < this.maxGameRooms)
                {
                    //register a new room
                    wr(ip + ":" + port + ": Register a new game room.");
                    r = new GameRoom();
                    r.serverVersion = this.version;
                    r.dunGenSeed = this.GetNewDunGenSeed();
                    r.id = this.gameRooms.Count;
                    r.eds3server = this;
                    this.gameRooms.Add(r);

                    //set player info
                    p.roomId = this.gameRooms.IndexOf(r);
                    r.players.Add(p.ip, p);

                    p.id = r.players[p.ip].GetHashCode();
                    p.spawnId = r.GetSpawnId();

                    //register player with room
                    this.players.Add(p.ip, r);
                }
                else
                {
                    wr(ip + ":" + port + ": Max rooms reached.");
                    r = null;
                    p.roomId = -1;
                    p.id = -1;
                    p.spawnId = -1;
                }
            }
            else
            {
                //get existing room
                wr(ip + ":" + port + ": Get existing game room.");
                r = gameRooms[id];

                //set player info
                p.roomId = this.gameRooms.IndexOf(r);
                r.players.Add(p.ip, p);
                p.id = r.players[p.ip].GetHashCode();
                p.spawnId = r.GetSpawnId();

                this.players.Add(p.ip, r);
            }

            if (r != null)
            {
                wr(ip + ":" + port + ": Assigning player to room: " + p.roomId + " player id: " + p.id);
                return true;
            }
            else
            {
                return false;
            }
        }

        private bool IsNewPlayer(string ip)
        {
            if (this.players != null)
            {
                if (this.players.ContainsKey(ip))
                {
                    return false;
                }
                else
                {
                    return true;
                }
            }
            else
            {
                this.players = new Dictionary<string, GameRoom>();
                return true;
            }
        }

        private bool IsExistingPlayer(string ip)
        {
            return !IsNewPlayer(ip);
        }

        public void ProcessCmd(string cmd)
        {
            wr("server processCmd: " + cmd);
            if (cmd != null)
            {
                cmd = cmd.ToLower().Trim();

                if (cmd.Equals("help"))
                {
                    wr("q: Quit server.");
                    wr("netstats: Print general network stats.");
                    wr("listrooms: Lists current game rooms.");
                    wr("listplayers-[game room index]: List players in the given game room index.");
                    wr("listlagplayers-[game room index]: List network lag players in the given game room index.");
                }
                else if (cmd.Equals("q"))
                {
                    wr("Stopping server...");
                    this.Stop();

                }
                else if (cmd.Equals("netstats"))
                {
                    wr("TotalNetworkDataIn: " + this.totalDataIn);
                    wr("TotalNetworkMsgsIn: " + this.totalMsgsIn);
                    wr("TotalNetworkDataOut: " + this.totalDataOut);
                    wr("TotalNetworkMsgsOut: " + this.totalMsgsOut);

                }
                else if (cmd.Equals("listrooms"))
                {
                    int len = this.gameRooms.Count;
                    wr("Scanning " + len + " rooms...");
                    for (int i = 0; i < len; i++)
                    {
                        wr("RoomIdx: " + (i + 1));
                        wr(this.gameRooms[i].ToString());
                    }

                }
                else if (cmd.IndexOf("listplayers") != -1)
                {
                    string[] s = null;
                    if (cmd.IndexOf("-") != -1)
                    {
                        s = cmd.Split('-');
                        int idx = int.Parse(s[1]);
                        GameRoom r = this.gameRooms[idx];
                        Player p = null;
                        int cnt = 0;
                        if (r != null)
                        {
                            foreach (string key in r.players.Keys)
                            {
                                p = r.players[key];
                                wr("PlayerIdx: " + (cnt + 1));
                                wr(p.ToString());
                                cnt++;
                            }
                        }
                    }
                    else
                    {
                        wr("No room specified...");
                    }

                }
                else if (cmd.IndexOf("listlagplayers") != -1)
                {
                    string[] s = null;
                    if (cmd.IndexOf("-") != -1)
                    {
                        s = cmd.Split('-');
                        int idx = int.Parse(s[1]);
                        GameRoom r = this.gameRooms[idx];
                        int cnt = 0;
                        if (r != null)
                        {
                            foreach (Player p in r.lagPlayers)
                            {
                                wr("LagPlayerIdx: " + (cnt + 1));
                                wr(p.ToString());
                                cnt++;
                            }
                        }
                    }
                    else
                    {
                        wr("No room specified...");
                    }

                }
            }
        }
        #endregion

        #region network_methods
        public void run()
        {
            StartListening();
        }

        //Begin listening for connection requests.
        private void StartListening()
        {
            try
            {
                wr("StartListening on: " + this.localIpAddress + ":" + this.port);
                this.endPoint = new IPEndPoint(IPAddress.Parse(this.localIpAddress), this.port);
                this.serverEndPoint = (EndPoint)this.endPoint;

                this.serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.IP);
                this.serverSocket.SendTimeout = this.timeout;
                this.serverSocket.ReceiveTimeout = this.timeout;
                this.serverSocket.Bind(this.serverEndPoint);

                wr("Listening on port: " + this.port + " Local IP: " + this.endPoint.Address);
                byte[] receiveData = new byte[this.inBufferSize];
                IPEndPoint client = new IPEndPoint(IPAddress.Any, 0);
                EndPoint clientEndPoint = (EndPoint)client;

                while (!this.shuttingDown)
                {
                    try
                    {
                        //Get next message block
                        int tlen = this.serverSocket.ReceiveFrom(receiveData, ref clientEndPoint);

                        //Handle next message block in a thread
                        if (ASYNC_ON)
                        {
                            AsyncNetworkReadCallBack(receiveData, tlen, clientEndPoint);
                        }
                        else
                        {
                            NetworkReadCallBack(receiveData, tlen, clientEndPoint);
                        }
                    }
                    catch (Exception e)
                    {
                        //ignore and keep listening
                    }

                    if (this.shuttingDown || !this.serverSocket.IsBound)
                    {
                        break;
                    }
                }

            }
            catch (Exception e)
            {
                if (!this.shuttingDown)
                {
                    wr(e.Message);
                    PrintStackTrace(e.StackTrace);
                }
            }
        }

        //Stop the service and shut it down.
        public void Stop()
        {
            this.CloseAllClients(true);

            try
            {
                Thread.Sleep(3000);
            }
            catch(Exception e)
            {
                
            }

            this.shuttingDown = true;

            try
            {
                //this.serverSocket.Close(1);
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
            }
        }

        //Start the service listening.
        public void Start()
        {
            wr("Starting: " + title + " Version: " + version);
            ThreadStart ts = new ThreadStart(StartListening);
            Thread trd = new Thread(ts);
            trd.Start();
        }

        public void CloseAllClients(bool b)
        {
            this.closeAllClients = b;
        }

        public void AsyncNetworkReadCallBack(byte[] dataIn, int tlen, EndPoint clientEndPoint)
        {
            try
            {
                lock (this)
                {
                    this.ancDataIn = dataIn;
                    this.ancLen = tlen;
                    this.ancClientEndPoint = clientEndPoint;
                }

                ThreadStart ts = new ThreadStart(NetworkReadCallBack);
                Thread trd = new Thread(ts);
                trd.Start();
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
            }
        }

        public void AsyncNetworkWriteCallBack(byte[] dataOut, int tlen, EndPoint clientEndPoint)
        {
            try
            {
                lock (this)
                {
                    this.ancDataOut = dataOut;
                    this.ancLen = tlen;
                    this.ancClientEndPoint = clientEndPoint;
                }

                ThreadStart ts = new ThreadStart(NetworkWriteCallBack);
                Thread trd = new Thread(ts);
                trd.Start();
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
            }
        }

        public void NetworkReadCallBack(byte[] dataIn, int tlen, EndPoint clientEndPoint)
        {
            lock (this)
            {
                this.ancDataIn = dataIn;
                this.ancLen = tlen;
                this.ancClientEndPoint = clientEndPoint;
            }

            this.NetworkReadCallBack();
        }

        public void NetworkReadCallBack()
        {
            //Main client request handler
            wr("");
            wr("");
            wr("NetworkReadCallBack called ");

            try
            {
                int tlen = this.ancLen;
                EndPoint clientEndPoint = this.ancClientEndPoint;
                IPEndPoint endPoint = ((IPEndPoint)clientEndPoint);

                if (this.shuttingDown || !this.serverSocket.IsBound)
                {
                    return;
                }

                this.totalDataIn += tlen;
                this.totalMsgsIn++;

                byte[] dataIn = this.ancDataIn;
                IPAddress ipAddress = endPoint.Address;
                string ip = ipAddress.ToString();
                int port = endPoint.Port;

                if (FORCE_PORT_USE)
                {
                    ip += ":" + port;
                }

                Player p = null;
                GameRoom r = null;

                wr("Found request id: " + dataIn[0]);
                if (dataIn[0] == (int)Eds3ServerMode.SERVER_RECEIVE_REQUEST_DUNGEN_SEED)
                {
                    wr(ip + ":" + port + ": Receiving SERVER_RECEIVE_REQUEST_DUNGEN_SEED msg.");

                    int dunGenSeed = -1;
                    string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                    long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                    long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                    MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                    wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                    if (IsNewPlayer(ip))
                    {
                        p = new Player(ip);
                        p.ip = ip;
                        p.port = port;
                        p.ipAddress = ipAddress;
                        p.clientVersion = version;
                        this.AssignGameRoom(p);
                    }

                    try
                    {
                        r = this.players[ip];
                    }
                    catch (Exception ex1)
                    {
                        PrintStackTrace(ex1.StackTrace);
                        r = null;
                    }

                    if (r != null)
                    {
                        p = r.players[ip];
                        if (p != null)
                        {
                            p.port = port;
                            p.ipAddress = ipAddress;
                            p.clientVersion = version;
                            dunGenSeed = r.dunGenSeed;

                            long start = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            p.lastCommTimeMs = iSentOn;
                            p.LogMsgStats(msgStats);
                            r.UpdateGameRoomStats();
                            r.UpdateLaggedInactivePlayers();
                            long stop = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            long diff = stop - start;
                            wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                        }
                        else
                        {
                            p = null;
                            dunGenSeed = DUN_GEN_SEED_ERROR;
                            wr(ip + ":" + port + ": Message chain dying...");
                        }
                    }
                    else
                    {
                        p = null;
                        dunGenSeed = DUN_GEN_SEED_ERROR;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }

                    wr(ip + ":" + port + ": ServerVersion: " + this.version);
                    wr(ip + ":" + port + ": ClientVersion: " + version);
                    wr(ip + ":" + port + ": DunGenSeed: " + dunGenSeed);
                    wr(ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3ServerMode.SERVER_SEND_REQUESTED_DUNGEN_SEED, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] data = Converter.toByte(dunGenSeed, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + data.Length + sentOn.Length;

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(data, 0, dataOut, (msgId.Length + msgVer.Length), data.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length + data.Length), sentOn.Length);

                    wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_DUNGEN_SEED msg");
                    NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);

                }
                else if (dataIn[0] == (int)Eds3ServerMode.SERVER_RECEIVE_REQUEST_SPAWN_ID)
                {
                    wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_SPAWN_ID msg.");

                    int spawnId = -1;
                    string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                    long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                    long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                    MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                    wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                    try
                    {
                        r = this.players[ip];
                    }
                    catch (Exception ex1)
                    {
                        PrintStackTrace(ex1.StackTrace);
                        r = null;
                    }

                    if (r != null)
                    {
                        p = r.players[ip];
                        if (p != null)
                        {
                            p.port = port;
                            p.ipAddress = ipAddress;
                            spawnId = p.spawnId;

                            long start = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            p.lastCommTimeMs = iSentOn;
                            p.LogMsgStats(msgStats);
                            r.UpdateGameRoomStats();
                            r.UpdateLaggedInactivePlayers();
                            long stop = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            long diff = stop - start;
                            wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                        }
                        else
                        {
                            p = null;
                            spawnId = SPAWN_ID_ERROR;
                            wr(ip + ":" + port + ": Message chain dying...");
                        }
                    }
                    else
                    {
                        p = null;
                        spawnId = SPAWN_ID_ERROR;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }

                    wr(ip + ":" + port + ": ServerVersion: " + this.version);
                    wr(ip + ":" + port + ": ClientVersion: " + version);
                    wr(ip + ":" + port + ": SpawnId/PlayerId: " + spawnId);
                    wr(ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3ServerMode.SERVER_SEND_REQUESTED_SPAWN_ID, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] data = Converter.toByte(spawnId, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + data.Length + sentOn.Length;

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(data, 0, dataOut, (msgId.Length + msgVer.Length), data.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length + data.Length), sentOn.Length);

                    wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_SPAWN_ID msg");
                    NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);

                }
                else if (dataIn[0] == (int)Eds3ServerMode.SERVER_RECEIVE_REQUEST_DROP_CLIENT)
                {
                    wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_DROP_CLIENT msg.");

                    int resId = 0;
                    string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                    long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                    long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                    MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                    wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                    try
                    {
                        r = this.players[ip];
                    }
                    catch (Exception ex1)
                    {
                        PrintStackTrace(ex1.StackTrace);
                        r = null;
                    }

                    if (r != null)
                    {
                        p = r.players[ip];
                        if (p != null)
                        {
                            p.port = port;
                            p.ipAddress = ipAddress;

                            r.players.Remove(ip);

                            if(r.lagPlayers.Contains(p))
                            {
                                r.lagPlayers.Remove(p);
                            }

                            if (r.inactivePlayers.Contains(p))
                            {
                                r.inactivePlayers.Remove(p);
                            }

                            this.players.Remove(ip);
                            if (!r.freeSpawns.Contains(p.spawnId))
                            {
                                r.freeSpawns.Add(p.spawnId);
                            }
                            resId = 2;

                            long start = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            //p.lastCommTimeMs = iSentOn;
                            p.lastCommTimeMs = 0;

                            p.LogMsgStats(msgStats);
                            r.UpdateGameRoomStats();
                            r.UpdateLaggedInactivePlayers();

                            long stop = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            long diff = stop - start;
                            wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                        }
                        else
                        {
                            p = null;
                            this.players.Remove(ip);
                            resId = 1;
                            wr(ip + ":" + port + ": Message chain dying...");
                        }
                    }
                    else
                    {
                        p = null;
                        resId = 0;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }

                    wr(ip + ":" + port + ": ServerVersion: " + this.version);
                    wr(ip + ":" + port + ": ClientVersion: " + version);
                    wr(ip + ":" + port + ": ResultId: " + resId);
                    wr(ip + ":" + port + ": SentOn: " + iSentOn);
                    wr(ip + ":" + port + ": ReceivedOn: " + recvOn);

                    byte[] msgId = new byte[] { Converter.toByte((int)Eds3ServerMode.SERVER_SEND_REQUESTED_DROP_CLIENT, REVERSE_TO_BYTES)[3] };
                    byte[] msgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                    byte[] data = Converter.toByte(resId, REVERSE_TO_BYTES);
                    byte[] sentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                    int totalLen = msgId.Length + msgVer.Length + data.Length + sentOn.Length;

                    byte[] dataOut = new byte[totalLen];
                    System.Array.Copy(msgId, 0, dataOut, 0, msgId.Length);
                    System.Array.Copy(msgVer, 0, dataOut, msgId.Length, msgVer.Length);
                    System.Array.Copy(data, 0, dataOut, (msgId.Length + msgVer.Length), data.Length);
                    System.Array.Copy(sentOn, 0, dataOut, (msgId.Length + msgVer.Length + data.Length), sentOn.Length);

                    wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_DROP_CLIENT msg");
                    NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);

                }
                else if (dataIn[0] == (int)Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1)
                {
                    wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L1 msg.");

                    string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                    long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                    long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                    MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                    wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                    int playerId = Converter.toInt(new byte[] { dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                    int objsLen = Converter.toInt(new byte[] { dataIn[17], dataIn[18], dataIn[19], dataIn[20] }, REVERSE_FROM_BYTES);
                    byte[] objs = new byte[objsLen];

                    byte[] dataOut;
                    byte b1;
                    byte b2;

                    try
                    {
                        System.Array.Copy(dataIn, 21, objs, 0, objs.Length);
                    }
                    catch (Exception ex2)
                    {
                        b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD, REVERSE_TO_BYTES)[3];

                        if (this.closeAllClients)
                        {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                        }
                        else
                        {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                            b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.WAITING)[3];
                        }
                        dataOut = new byte[] { b1, b2 };
                        NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                    }

                    try
                    {
                        r = this.players[ip];
                    }
                    catch (Exception ex1)
                    {
                        PrintStackTrace(ex1.StackTrace);
                        r = null;
                    }

                    if (r != null)
                    {
                        p = r.players[ip];
                        if (p != null)
                        {
                            p.port = port;
                            p.ipAddress = ipAddress;
                            p.objPublish = objs;

                            long start = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            p.lastCommTimeMs = iSentOn;
                            p.LogMsgStats(msgStats);
                            r.UpdateGameRoomStats();
                            r.UpdateLaggedInactivePlayers();
                            long stop = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            long diff = stop - start;
                            wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                        }
                        else
                        {
                            p = null;
                            wr(ip + ":" + port + ": Message chain dying...");
                        }
                    }
                    else
                    {
                        p = null;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }

                    if (r != null && p != null)
                    {
                        if (r.inactivePlayers.Count > 0)
                        {
                            wr(ip + ":" + port + ": Found inactive players.");
                            Player iap = r.inactivePlayers[0];
                            r.inactivePlayers.RemoveAt(0);
                            GameObjectWrapperDefault iapGow = new GameObjectWrapperDefault();
                            iapGow.SetPlayerId(iap.spawnId);
                            iapGow.SetMode((int)GameObjectWrapperDefault.GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR);
                            iapGow.SetId(iap.id);
                            byte[] iapB = iapGow.Serialize();

                            byte[] nop = null;
                            Dictionary<string, Player>.KeyCollection pks = r.players.Keys;
                            Player tiap;
                            foreach (string pk in pks)
                            {
                                tiap = r.players[pk];
                                if (tiap.objPublish == null)
                                {
                                    tiap.objPublish = iapB;
                                }
                                else
                                {
                                    nop = new byte[p.objPublish.Length + iapB.Length];
                                    System.Array.Copy(p.objPublish, 0, nop, 0, p.objPublish.Length);
                                    System.Array.Copy(iapB, 0, nop, p.objPublish.Length, iapB.Length);
                                    tiap.objPublish = nop;
                                }
                            }
                        }
                    
                        wr(ip + ":" + port + ": ServerVersion: " + this.version);
                        wr(ip + ":" + port + ": ClientVersion: " + version);
                        wr(ip + ":" + port + ": SpawnId/PlayerId: " + p.spawnId);
                        wr(ip + ":" + port + ": Object Id: " + p.id);
                        wr(ip + ":" + port + ": SentOn: " + iSentOn);
                        wr(ip + ":" + port + ": ReceivedOn: " + recvOn);
                        wr(ip + ":" + port + ": In SpawnId/PlayerId: " + playerId);
                        wr(ip + ":" + port + ": Serialized Objects Length: " + objs.Length);
                        wr(ip + ":" + port + ": Serialized Object Count: " + objs.Length / this.gowStatic.GetTotalLength());

                        byte[] tmsgId;
                        byte[] tmsgVer;
                        byte[] tsentOn;
                        byte[] tpId;
                        byte[] tobjs;
                        byte[] tobjsLen;
                        int totalLen;
                        byte[] lb;
                        Player tp;

                        foreach (string key in r.players.Keys)
                        {
                            tp = r.players[key];
                            if (tp != null && tp.spawnId != p.spawnId && tp.objPublish != null)
                            {
                                tmsgId = new byte[] { Converter.toByte((int)Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1, REVERSE_TO_BYTES)[3] };
                                tmsgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                                tsentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                                tpId = Converter.toByte(tp.spawnId, REVERSE_TO_BYTES);
                                tobjs = tp.objPublish;
                                tobjsLen = Converter.toByte(tobjs.Length, REVERSE_TO_BYTES);
                                totalLen = tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length + tobjs.Length + tobjsLen.Length;

                                dataOut = new byte[totalLen];
                                System.Array.Copy(tmsgId, 0, dataOut, 0, tmsgId.Length);
                                System.Array.Copy(tmsgVer, 0, dataOut, tmsgId.Length, tmsgVer.Length);
                                System.Array.Copy(tsentOn, 0, dataOut, (tmsgId.Length + tmsgVer.Length), tsentOn.Length);
                                System.Array.Copy(tpId, 0, dataOut, (tmsgId.Length + tmsgVer.Length + tsentOn.Length), tpId.Length);
                                System.Array.Copy(tobjsLen, 0, dataOut, (tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length), tobjsLen.Length);
                                System.Array.Copy(tobjs, 0, dataOut, (tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length + tobjsLen.Length), tobjs.Length);

                                wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L1 msg for SpawnId/PlayerId: " + tp.spawnId + " to " + p.spawnId);
                                if (this.commH != null)
                                {
                                    dataOut = this.commH.HandleObjPublication(dataOut, dataOut.Length, this);
                                }

                                if(tp != null && tp.isInactive) 
                                {
                                    wr(ip + ":" + port + ": Dropping player " + tp.spawnId + " with IP: " + tp.ip + " and ID: " + tp.id + ".");
                                    r.players.Remove(tp.ip);
                                    this.players.Remove(tp.ip);
                                    
                                    if(r.inactivePlayers.Contains(tp)) {
                                        r.inactivePlayers.Remove(tp);
                                    }
                                    
                                    if(r.lagPlayers.Contains(tp)) {
                                        r.lagPlayers.Remove(tp);
                                    }

                                    if (!r.freeSpawns.Contains(p.spawnId)) {
                                        r.freeSpawns.Add(p.spawnId);
                                    }
                                }
                            
                                NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                            }
                            else
                            {
                                if (tp == null)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to player no player entry found.");
                                }
                                else if (tp.spawnId == p.spawnId)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + ", " + p.spawnId);
                                }
                                else if (tp.objPublish == null)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to player no objects array to publish.");
                                }
                            }
                        }

                        if (this.commH != null)
                        {
                            lb = this.commH.HandleProcessServerGame(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), this);
                            if (lb != null && lb.Length > 0)
                            {
                                wr(ip + ":" + port + ": Sending HandleProcessServerGame msg");
                                NetworkWriteCallBack(lb, lb.Length, clientEndPoint);
                            }
                        }

                        if (SERVER_DRIVEN_MODE_ON)
                        {
                            b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD, REVERSE_TO_BYTES)[3];

                            if (this.closeAllClients)
                            {
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                            }
                            else
                            {
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.WAITING)[3];
                            }
                            dataOut = new byte[] { b1, b2 };
                            NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                        }
                        else
                        {
                            if (this.closeAllClients)
                            {
                                b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)[3];
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                                dataOut = new byte[] { b1, b2 };
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                                NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                            }
                        }
                    }
                    else 
                    {
                        b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)[3];
                        b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                        dataOut = new byte[] { b1, b2 };
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                        NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                    }
                }
                else if (dataIn[0] == (int)Eds3ServerMode.SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2)
                {
                    wr(ip + ":" + port + ": Found SERVER_RECEIVE_REQUEST_OBJ_PUBLISH_L2 msg.");

                    string version = Converter.toString(new byte[] { dataIn[1], dataIn[2], dataIn[3], dataIn[4] }, REVERSE_FROM_BYTES);
                    long iSentOn = Converter.toLong(new byte[] { dataIn[5], dataIn[6], dataIn[7], dataIn[8], dataIn[9], dataIn[10], dataIn[11], dataIn[12] }, REVERSE_FROM_BYTES);
                    long recvOn = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                    MsgStats msgStats = new MsgStats(iSentOn, recvOn);
                    wr(ip + ":" + port + ": XportTime: " + msgStats.diff + " ms");

                    int playerId = Converter.toInt(new byte[] { dataIn[13], dataIn[14], dataIn[15], dataIn[16] }, REVERSE_FROM_BYTES);
                    long iX = Converter.toLong(new byte[] { dataIn[17], dataIn[18], dataIn[19], dataIn[20], dataIn[21], dataIn[22], dataIn[23], dataIn[24] }, REVERSE_FROM_BYTES);
                    long iY = Converter.toLong(new byte[] { dataIn[25], dataIn[26], dataIn[27], dataIn[28], dataIn[29], dataIn[30], dataIn[31], dataIn[32] }, REVERSE_FROM_BYTES);
                    long iZ = Converter.toLong(new byte[] { dataIn[33], dataIn[34], dataIn[35], dataIn[36], dataIn[37], dataIn[38], dataIn[39], dataIn[40] }, REVERSE_FROM_BYTES);
                    //long iRot = Converter.toLong(new byte[] { dataIn[41], dataIn[42], dataIn[43], dataIn[44], dataIn[45], dataIn[46], dataIn[47], dataIn[48] }, REVERSE_FROM_BYTES);
                    int objsLen = Converter.toInt(new byte[] { dataIn[41], dataIn[42], dataIn[43], dataIn[44] }, REVERSE_FROM_BYTES);
                    byte[] objs = new byte[objsLen];

                    byte[] dataOut;
                    byte b1;
                    byte b2;

                    try
                    {
                        System.Array.Copy(dataIn, 45, objs, 0, objs.Length);
                    }
                    catch (Exception ex2)
                    {
                        b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD, REVERSE_TO_BYTES)[3];

                        if (this.closeAllClients)
                        {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                            b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                        }
                        else
                        {
                            wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                            b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.WAITING)[3];
                        }
                        dataOut = new byte[] { b1, b2 };
                        NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                    }

                    try
                    {
                        r = this.players[ip];
                    }
                    catch (Exception ex1)
                    {
                        PrintStackTrace(ex1.StackTrace);
                        r = null;
                    }

                    if (r != null)
                    {
                        p = r.players[ip];
                        if (p != null)
                        {
                            p.port = port;
                            p.ipAddress = ipAddress;
                            p.objPublish = objs;

                            p.x = (float)iX;
                            p.y = (float)iY;
                            p.z = (float)iZ;

                            long start = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            p.lastCommTimeMs = iSentOn;
                            p.LogMsgStats(msgStats);
                            r.UpdateGameRoomStats();
                            r.UpdateLaggedInactivePlayers();
                            long stop = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();
                            long diff = stop - start;
                            wr(ip + ":" + port + ": Player maintenance took: " + diff + " ms");
                        }
                        else
                        {
                            p = null;
                            wr(ip + ":" + port + ": Message chain dying...");
                        }
                    }
                    else
                    {
                        p = null;
                        wr(ip + ":" + port + ": Message chain dying...");
                    }

                    if (r != null && p != null)
                    {
                        if (r.inactivePlayers.Count > 0)
                        {
                            Player iap = r.inactivePlayers[0];
                            GameObjectWrapperDefault iapGow = new GameObjectWrapperDefault();
                            iapGow.SetPlayerId(iap.spawnId);
                            iapGow.SetMode((int)GameObjectWrapperDefault.GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR);
                            iapGow.SetId(iap.id);
                            byte[] iapB = iapGow.Serialize();

                            byte[] nop = null;
                            Dictionary<string, Player>.KeyCollection pks = r.players.Keys;
                            Player tiap;
                            foreach (string pk in pks)
                            {
                                tiap = r.players[pk];
                                if (tiap.objPublish == null)
                                {
                                    tiap.objPublish = iapB;
                                }
                                else
                                {
                                    nop = new byte[p.objPublish.Length + iapB.Length];
                                    System.Array.Copy(p.objPublish, 0, nop, 0, p.objPublish.Length);
                                    System.Array.Copy(iapB, 0, nop, p.objPublish.Length, iapB.Length);
                                    tiap.objPublish = nop;
                                }
                            }
                        }
                    
                        wr(ip + ":" + port + ": ServerVersion: " + this.version);
                        wr(ip + ":" + port + ": ClientVersion: " + version);
                        wr(ip + ":" + port + ": SpawnId/PlayerId: " + p.spawnId);
                        wr(ip + ":" + port + ": Object Id: " + p.id);
                        wr(ip + ":" + port + ": SentOn: " + iSentOn);
                        wr(ip + ":" + port + ": ReceivedOn: " + recvOn);
                        wr(ip + ":" + port + ": In SpawnId/PlayerId: " + playerId);
                        wr(ip + ":" + port + ": In X,Y,Z: " + iX + ", " + iY + ", " + iZ);
                        wr(ip + ":" + port + ": Serialized Objects Length: " + objs.Length);
                        wr(ip + ":" + port + ": Serialized Object Count: " + objs.Length / this.gowStatic.GetTotalLength());

                        byte[] tmsgId;
                        byte[] tmsgVer;
                        byte[] tsentOn;
                        byte[] tpId;
                        byte[] tpIp;
                        byte[] tobjs;
                        byte[] tobjsLen;
                        int totalLen;
                        byte[] lb;
                        string ttt;
                        Player tp = null;
                        dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                        int dataOutUsed = 0;
                        int dataOutStart = 0;
                        int cnt = 0;
                        Player prevTp = null;

                        foreach (string key in r.players.Keys)
                        {
                            tp = r.players[key];
                            if (tp != null && tp.spawnId != p.spawnId && tp.objPublish != null)
                            {
                                prevTp = tp;
                                if (Eds3Server.OBJ_PUBLISH_L2_DIST_LIMIT_ON && Utils.SqrDist(iX, iY, iZ, tp) > Eds3Server.OBJ_PUBLISH_L2_DIST_MAX_SQR)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + " from " + p.spawnId + " too distant.");
                                    continue;
                                }
                                tmsgId = new byte[] { Converter.toByte((int)Eds3ServerMode.SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2, REVERSE_TO_BYTES)[3] };
                                tmsgVer = Converter.toByte(this.version, REVERSE_TO_BYTES);
                                tsentOn = Converter.toByte(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), REVERSE_TO_BYTES);
                                tpId = Converter.toByte(tp.spawnId, REVERSE_TO_BYTES);
                                ttt = tp.ip;
                                tpIp = Converter.toByte(ttt);
                                if (tpIp.Length < IP_SER_LENGTH)
                                {
                                    ttt = Utils.PadRight(ttt, IP_SER_LENGTH - ttt.Length);
                                    tpIp = Converter.toByte(ttt, REVERSE_TO_BYTES);
                                }

                                tobjs = tp.objPublish;
                                tobjsLen = Converter.toByte(tobjs.Length, REVERSE_TO_BYTES);

                                wr("===================tpIpLen: " + tpIp.Length + " tobjsLen: " + tobjs.Length + " SpawnId/PlayerId: " + tp.spawnId);
                                totalLen = tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length + tpIp.Length + tobjsLen.Length + tobjs.Length;

                                if ((dataOutUsed + totalLen) >= dataOut.Length)
                                {
                                    wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg for SpawnId/PlayerId: " + tp.spawnId + " FrameCount: " + cnt + " to " + p.spawnId);
                                    if (this.commH != null)
                                    {
                                        dataOut = this.commH.HandleObjPublication(dataOut, dataOut.Length, this);
                                    }

                                    if(tp != null && tp.isInactive) 
                                    {
                                        wr(ip + ":" + port + ": Dropping player " + tp.spawnId + " with IP: " + tp.ip + " and ID: " + tp.id + ".");
                                        r.players.Remove(tp.ip);
                                        this.players.Remove(tp.ip);

                                        if(r.inactivePlayers.Contains(tp)) {
                                            r.inactivePlayers.Remove(tp);
                                        }

                                        if(r.lagPlayers.Contains(tp)) {
                                            r.lagPlayers.Remove(tp);
                                        }

                                        if (!r.freeSpawns.Contains(p.spawnId)) {
                                            r.freeSpawns.Add(p.spawnId);
                                        }
                                    }
                            
                                    NetworkWriteCallBack(dataOut, dataOutUsed, clientEndPoint);

                                    dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                                    dataOutStart = 0;
                                    dataOutUsed = 0;
                                    cnt = 0;
                                }

                                dataOutStart = dataOutUsed;
                                dataOutUsed += totalLen;
                                System.Array.Copy(tmsgId, 0, dataOut, dataOutStart, tmsgId.Length);
                                System.Array.Copy(tmsgVer, 0, dataOut, (dataOutStart + tmsgId.Length), tmsgVer.Length);
                                System.Array.Copy(tsentOn, 0, dataOut, (dataOutStart + tmsgId.Length + tmsgVer.Length), tsentOn.Length);
                                System.Array.Copy(tpId, 0, dataOut, (dataOutStart + tmsgId.Length + tmsgVer.Length + tsentOn.Length), tpId.Length);
                                System.Array.Copy(tpIp, 0, dataOut, (dataOutStart + tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length), tpIp.Length);
                                System.Array.Copy(tobjsLen, 0, dataOut, (dataOutStart + tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length + tpIp.Length), tobjsLen.Length);
                                System.Array.Copy(tobjs, 0, dataOut, (dataOutStart + tmsgId.Length + tmsgVer.Length + tsentOn.Length + tpId.Length + tpIp.Length + tobjsLen.Length), tobjs.Length);
                                cnt++;
                            }
                            else
                            {
                                if (tp == null)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to player no player entry found.");
                                }
                                else if (tp.spawnId == p.spawnId)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to source player: " + tp.spawnId + ", " + p.spawnId);
                                }
                                else if (tp.objPublish == null)
                                {
                                    wr(ip + ":" + port + ": Skipping msg to player no objects array to publish.");
                                }
                            }
                        }

                        if (dataOutUsed > 0)
                        {
                            if (prevTp != null)
                            {
                                wr(prevTp + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg for SpawnId/PlayerId: " + prevTp.spawnId);
                            }
                            else
                            {
                                wr(prevTp + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg");
                            }

                            wr(ip + ":" + port + ": Sending SERVER_SEND_REQUESTED_OBJ_PUBLISH_L2 msg for SpawnId/PlayerId: " + prevTp.spawnId);
                            if (this.commH != null)
                            {
                                dataOut = this.commH.HandleObjPublication(dataOut, dataOut.Length, this);
                            }

                            if (prevTp != null && prevTp.isInactive)
                            {
                                wr(ip + ":" + port + ": Dropping player " + prevTp.spawnId + " with IP: " + prevTp.ip + " and ID: " + prevTp.id + ".");
                                r.players.Remove(prevTp.ip);
                                this.players.Remove(prevTp.ip);

                                if (r.inactivePlayers.Contains(prevTp))
                                {
                                    r.inactivePlayers.Remove(prevTp);
                                }

                                if (r.lagPlayers.Contains(prevTp))
                                {
                                    r.lagPlayers.Remove(prevTp);
                                }

                                if (!r.freeSpawns.Contains(p.spawnId))
                                {
                                    r.freeSpawns.Add(p.spawnId);
                                }
                            }

                            NetworkWriteCallBack(dataOut, dataOutUsed, clientEndPoint);

                            dataOut = new byte[Eds3Server.DEFAULT_OUT_BUFFER];
                            dataOutStart = 0;
                            dataOutUsed = 0;
                        }

                        if (this.commH != null)
                        {
                            lb = this.commH.HandleProcessServerGame(System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds(), this);
                            if (lb != null && lb.Length > 0)
                            {
                                wr(ip + ":" + port + ": Sending HandleProcessServerGame msg");
                                NetworkWriteCallBack(lb, lb.Length, clientEndPoint);
                            }
                        }

                        if (SERVER_DRIVEN_MODE_ON)
                        {
                            b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD, REVERSE_TO_BYTES)[3];

                            if (this.closeAllClients)
                            {
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                            }
                            else
                            {
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to WAITING");
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.WAITING)[3];
                            }
                            dataOut = new byte[] { b1, b2 };
                            NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                        }
                        else
                        {
                            if (this.closeAllClients)
                            {
                                b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)[3];
                                b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                                dataOut = new byte[] { b1, b2 };
                                wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                                NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                            }
                        }
                    }
                    else
                    {
                        b1 = Converter.toByte((int)Eds3ServerMode.SERVER_SEND_SET_MODE_CMD)[3];
                        b2 = Converter.toByte((int)Eds3Client.Eds3ClientMode.CLIENT_SEND_REQUEST_DROP_CLIENT)[3];
                        dataOut = new byte[] { b1, b2 };
                        wr(ip + ":" + port + ": Sending SERVER_SEND_SET_MODE_CMD msg to CLIENT_SEND_REQUEST_DROP_CLIENT");
                        NetworkWriteCallBack(dataOut, dataOut.Length, clientEndPoint);
                    }
                }
                else
                {
                    if (this.commH != null)
                    {
                        this.commH.HandleProcessModeUnknown(dataIn, this);
                    }
                }
            }
            catch (Exception e)
            {
                if (!this.shuttingDown)
                {
                    wr(e.Message);
                    PrintStackTrace(e.StackTrace);
                }
            }
        }

        public void NetworkWriteCallBack(byte[] dataOut, int tlen, EndPoint clientEndPoint)
        {
            lock (this)
            {
                this.ancDataOut = dataOut;
                this.ancLen = tlen;
                this.ancClientEndPoint = clientEndPoint;
            }
            this.NetworkWriteCallBack();
        }

        public void NetworkWriteCallBack()
        {
            //Retrieve the state object and the handler socket.
            wr("");
            wr("");
            wr("NetworkWriteCallBack");

            try
            {
                byte[] dataOut = null;
                int tlen = 0;
                EndPoint clientEndPoint = null;

                lock (this)
                {
                    dataOut = this.ancDataOut;
                    tlen = this.ancLen;
                    clientEndPoint = this.ancClientEndPoint;
                }

                if (this.shuttingDown || !this.serverSocket.IsBound)
                {
                    return;
                }

                wr("DataOutLen: " + dataOut.Length + " TLen: " + tlen);
                this.totalDataOut += tlen;
                this.totalMsgsOut++;
                serverSocket.SendTo(dataOut, clientEndPoint);
            }
            catch (Exception e)
            {
                wr(e.Message);
                PrintStackTrace(e.StackTrace);
            }
        }
        #endregion

        public void SetDebugOn(bool b)
        {
            this.debugOn = b;
        }
    }
}
