﻿using System;
using System.Collections.Generic;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3.unity
{
    public class GameObjectManagerDefault : GameObjectManager
    {
        private Dictionary<int, GameObjectWrapper> gameObjects;
        private Dictionary<int, GameObjectWrapper> networkGameObjects;
        private Dictionary<int, GameObjectWrapper> inNetworkGameObjects;
        private HandleGameObjectWrapperSerialization serH = null;
        private GameObjectWrapper playerChar = null;
        private Dictionary<string, int> nameToGameObjectId = null;

        public GameObjectManagerDefault()
        {
            this.playerChar = new GameObjectWrapperDefault();
            this.gameObjects = new Dictionary<int, GameObjectWrapper>();
            this.networkGameObjects = new Dictionary<int, GameObjectWrapper>();
            this.nameToGameObjectId = new Dictionary<string, int>();
            this.inNetworkGameObjects = new Dictionary<int, GameObjectWrapper>();
        }

        public Dictionary<int, GameObjectWrapper> GetGameObjects()
        {
            return this.gameObjects;
        }

        public Dictionary<int, GameObjectWrapper> GetNetworkGameObjects()
        {
            return this.networkGameObjects;
        }

        public Dictionary<int, GameObjectWrapper> GetInNetworkGameObjects()
        {
            return this.inNetworkGameObjects;
        }

        public Dictionary<string, int> GetNameToGameObjectId()
        {
            return this.nameToGameObjectId;
        }

        public GameObjectWrapper GetPlayerChar()
        {
            return this.playerChar;
        }

        public virtual void SetPlayerChar(GameObjectWrapper pc)
        {
            this.playerChar = pc;
        }

        public virtual HandleGameObjectWrapperSerialization GetSerializationHandler()
        {
            return this.serH;
        }

        public virtual void SetSerializationHandler(HandleGameObjectWrapperSerialization SerH)
        {
            this.serH = SerH;
        }

        public virtual void RegisterGameObjectName(string name, int id)
        {
            if (this.nameToGameObjectId.ContainsKey(name))
            {
                this.nameToGameObjectId.Remove(name);
            }
            this.nameToGameObjectId.Add(name, id);
        }

        public virtual void AddGameObject(GameObjectWrapper obj, bool isNetwork)
        {
            if (isNetwork)
            {
                if (this.networkGameObjects.ContainsKey(obj.GetId()))
                {
                    this.networkGameObjects.Remove(obj.GetId());
                }
                this.networkGameObjects.Add(obj.GetId(), obj);
            }
            else
            {
                if (this.gameObjects.ContainsKey(obj.GetId()))
                {
                    this.gameObjects.Remove(obj.GetId());
                }
                this.gameObjects.Add(obj.GetId(), obj);
            }
        }

        public void AddInNetworkGameObject(GameObjectWrapper obj)
        {
            if (this.inNetworkGameObjects.ContainsKey(obj.GetId()))
            {
                this.inNetworkGameObjects.Remove(obj.GetId());
            }
            this.inNetworkGameObjects.Add(obj.GetId(), obj);
        }

        public virtual byte[] SerializeNetworkGameObjects()
        {
            return this.SerializeDirect(this.networkGameObjects);
        }

        public virtual byte[] SerializeGameObjects()
        {
            return this.SerializeDirect(this.gameObjects);
        }

        public virtual byte[] Serialize(Dictionary<int, GameObjectWrapper> bIn)
        {
            List<GameObjectWrapper> bOut = new List<GameObjectWrapper>();
            bOut.AddRange(bIn.Values);
            return this.Serialize(bOut);
        }

        public virtual byte[] Serialize(List<GameObjectWrapper> bIn2)
        {
            List<GameObjectWrapper> bIn = new List<GameObjectWrapper>(bIn2);
            int len = bIn.Count;


            bool hasPlayerChar = false;
            if (this.playerChar != null)
            {
                hasPlayerChar = true;
                len += 1;
            }

            int off = 0;
            int[] offs = new int[len + 1];
            int[] lengths = new int[len];

            List<byte[]> pRet = new List<byte[]>();
            int totalLen = 0;
            byte[] b;
            byte[] ret;
            GameObjectWrapper gow;

            for (int i = 0; i < len; i++)
            {
                if (hasPlayerChar && i == 0)
                {
                    gow = this.playerChar;

                }
                else
                {
                    gow = bIn[i];

                }

                if (this.serH != null)
                {
                    gow = this.serH.BeforeSerializeGow(gow);
                }

                b = gow.Serialize();

                if (this.serH != null)
                {
                    b = this.serH.AfterSerializeGow(b, gow);
                }

                pRet.Add(b);

                if (this.serH != null)
                {
                    this.serH.SerializeIteration(gow, b, pRet, off);
                }

                pRet.Add(b);
                offs[i] = off;
                off += b.Length;
                lengths[i] = b.Length;
                totalLen += b.Length;
            }
            offs[offs.Length - 1] = totalLen;

            ret = new byte[totalLen];
            for (int i = 0; i < len; i++)
            {
                System.Array.Copy(pRet[i], 0, ret, offs[i], lengths[i]);
            }

            return ret;
        }

        public virtual byte[] SerializeDirect(Dictionary<int, GameObjectWrapper> bIn2)
        {
            Dictionary<int, GameObjectWrapper> bIn = new Dictionary<int, GameObjectWrapper>(bIn2);
            int len = bIn.Count;

            bool hasPlayerChar = false;
            if (this.playerChar != null)
            {
                hasPlayerChar = true;
                len += 1;
            }

            int off = 0;
            int totalLen = len * GameObjectWrapperDefault.TOTAL_LENGTH;
            byte[] b;
            byte[] ret = new byte[totalLen];
            GameObjectWrapper gow2;

            if (hasPlayerChar)
            {
                gow2 = this.playerChar;

                if (this.serH != null)
                {
                    gow2 = this.serH.BeforeSerializeGow(gow2);
                }

                b = gow2.Serialize();

                if (this.serH != null)
                {
                    b = this.serH.AfterSerializeGow(b, gow2);
                }

                System.Array.Copy(b, 0, ret, off, GameObjectWrapperDefault.TOTAL_LENGTH);

                if (this.serH != null)
                {
                    this.serH.SerializeIteration(gow2, b, ret, off);
                }

                off += b.Length;
            }

            foreach (GameObjectWrapper gow in bIn.Values)
            {
                gow2 = gow;

                if (this.serH != null)
                {
                    gow2 = this.serH.BeforeSerializeGow(gow2);
                }

                b = gow2.Serialize();

                if (this.serH != null)
                {
                    b = this.serH.AfterSerializeGow(b, gow2);
                }

                System.Array.Copy(b, 0, ret, off, GameObjectWrapperDefault.TOTAL_LENGTH);

                if (this.serH != null)
                {
                    this.serH.SerializeIteration(gow2, b, ret, off);
                }

                off += b.Length;
            }
            return ret;
        }

        public virtual List<GameObjectWrapper> Deserialize2ArrayList(byte[] bIn)
        {
            int len = bIn.Length;
            List<GameObjectWrapper> ret = new List<GameObjectWrapper>();
            GameObjectWrapper gow = null;
            int msgLen = GameObjectWrapperDefault.TOTAL_LENGTH;
            byte[] d = null;
            int pos = 0;

            if (this.playerChar != null)
            {
                d = new byte[msgLen];
                System.Array.Copy(bIn, pos, d, 0, msgLen);
                gow = new GameObjectWrapperDefault();

                if (this.serH != null)
                {
                    d = this.serH.BeforeDeserializeGow(d, gow);
                }

                gow.Deserialize(d);

                if (this.serH != null)
                {
                    gow = this.serH.AfterDeserializeGow(gow);
                }

                this.playerChar = gow;

                if (this.serH != null)
                {
                    this.serH.DeserializeIteration(gow, d, ret, pos);
                }

                pos += msgLen;
                len -= msgLen;
            }

            while (len >= msgLen)
            {
                d = new byte[msgLen];
                System.Array.Copy(bIn, pos, d, 0, msgLen);
                gow = new GameObjectWrapperDefault();

                if (this.serH != null)
                {
                    d = this.serH.BeforeDeserializeGow(d, gow);
                }

                gow.Deserialize(d);

                if (this.serH != null)
                {
                    gow = this.serH.AfterDeserializeGow(gow);
                }

                ret.Add(gow);

                if (this.serH != null)
                {
                    this.serH.DeserializeIteration(gow, d, ret, pos);
                }

                pos += msgLen;
                len -= msgLen;
            }

            Utils.wr("Found " + ret.Count + " game objects.");
            return ret;
        }

        public virtual Dictionary<int, GameObjectWrapper> Deserialize2Hashtable(byte[] bIn)
        {
            List<GameObjectWrapper> tmp = this.Deserialize2ArrayList(bIn);
            GameObjectWrapper gow = null;
            int len = tmp.Count;
            Dictionary<int, GameObjectWrapper> ret = new Dictionary<int, GameObjectWrapper>();

            for (int i = 0; i < len; i++)
            {
                gow = tmp[i];
                ret.Add(gow.GetId(), gow);
            }

            Utils.wr("Found " + ret.Count + " game objects.");
            return ret;
        }

        public virtual Dictionary<int, GameObjectWrapper> Deserialize2HashtableDirect(byte[] bIn)
        {
            int len = bIn.Length;
            Dictionary<int, GameObjectWrapper> ret = new Dictionary<int, GameObjectWrapper>();
            GameObjectWrapper gow = null;
            int msgLen = GameObjectWrapperDefault.TOTAL_LENGTH;
            byte[] d = null;
            int pos = 0;

            if (this.playerChar != null)
            {
                d = new byte[msgLen];
                System.Array.Copy(bIn, pos, d, 0, msgLen);
                gow = new GameObjectWrapperDefault();

                if (this.serH != null)
                {
                    d = this.serH.BeforeDeserializeGow(d, gow);
                }

                gow.Deserialize(d);

                if (this.serH != null)
                {
                    gow = this.serH.AfterDeserializeGow(gow);
                }

                this.playerChar = gow;

                if (this.serH != null)
                {
                    this.serH.DeserializeIteration(gow, d, ret, pos);
                }

                pos += msgLen;
                len -= msgLen;
            }

            while (len >= msgLen)
            {
                d = new byte[msgLen];
                System.Array.Copy(bIn, pos, d, 0, msgLen);
                gow = new GameObjectWrapperDefault();

                if (this.serH != null)
                {
                    d = this.serH.BeforeDeserializeGow(d, gow);
                }

                gow.Deserialize(d);

                if (this.serH != null)
                {
                    gow = this.serH.AfterDeserializeGow(gow);
                }

                ret.Add(gow.GetId(), gow);

                if (this.serH != null)
                {
                    this.serH.DeserializeIteration(gow, d, ret, pos);
                }

                pos += msgLen;
                len -= msgLen;
            }

            Utils.wr("Found " + ret.Count + " game objects.");
            return ret;
        }

        public virtual void DeserializeNetworkGameObjects(byte[] bIn)
        {
            this.networkGameObjects = this.Deserialize2HashtableDirect(bIn);
        }

        public virtual void DeserializeGameObjects(byte[] bIn)
        {
            this.gameObjects = this.Deserialize2HashtableDirect(bIn);
        }

        public virtual bool NetworkGameObjectsEquals(GameObjectManager gom)
        {
            GameObjectWrapper gow1 = null;
            GameObjectWrapper gow2 = null;

            foreach (int key in gom.GetNetworkGameObjects().Keys)
            {
                gow1 = gom.GetNetworkGameObjects()[key];
                gow2 = this.networkGameObjects[key];

                if (gow1.Equals(gow2) == false)
                {
                    return false;
                }
            }

            return true;
        }

        public virtual bool GameObjectsEquals(GameObjectManager gom)
        {
            GameObjectWrapper gow1 = null;
            GameObjectWrapper gow2 = null;

            foreach (int key in gom.GetGameObjects().Keys)
            {
                gow1 = gom.GetGameObjects()[key];
                gow2 = this.gameObjects[key];

                if (gow1.Equals(gow2) == false)
                {
                    return false;
                }
            }

            return true;
        }
    }
}
