﻿using System;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3.unity
{
    public class GameObjectWrapperDefault : GameObjectWrapper
    {
        public enum GameObjectWrapperMode
        {
            UPDATE,                     //0
            CREATE,                     //1
            DESTROY,                    //2
            NONE,                       //3
            ERROR,                      //4
            SPECIAL_1,
            SPECIAL_2,
            SPECIAL_3,
            SPECIAL_4,
            SPECIAL_5,
            SPECIAL_6,
            SPECIAL_7,
            SPECIAL_8,
            SPECIAL_9,
            SPECIAL_10,
            ACTIVATE,
            DEACTIVATE,
            UPDATE_POS,
            UPDATE_ROT,
            UPDATE_VEL,
            IGNORE,
            DESTROY_INACTIVE_PLAYER_CHAR
        }

        public enum GameObjectWrapperSerMode
        {
            L1,         //pos: x,y,z        rot: w,x,y,z
            L2          //pos: x,y,z        rot: w,x,y,z        vel: x,y,z
        }

        public static GameObjectWrapperSerMode OBJ_SER_MODE = GameObjectWrapperSerMode.L2;              //x  y  z  rx ry rz rw id pid mode - 10
        public static int[] BINARY_LENGTHS = (OBJ_SER_MODE == GameObjectWrapperSerMode.L1) ? (new int[] { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8 }) : (new int[] { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8 });
        public static int[] BINARY_OFFSETS = (OBJ_SER_MODE == GameObjectWrapperSerMode.L1) ? (new int[] { 0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 52 }) : (new int[] { 0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 64 });
        public static int TOTAL_LENGTH = (OBJ_SER_MODE == GameObjectWrapperSerMode.L1) ? (52) : (64);
        public static bool GEN_RANDOM_ATTRS = true;

        private float x = 0.0f;
        private float y = 0.0f;
        private float z = 0.0f;
        private float rotX = 0.0f;
        private float rotY = 0.0f;
        private float rotZ = 0.0f;
        private float rotW = 0.0f;
        private float velX = 0.0f;
        private float velY = 0.0f;
        private float velZ = 0.0f;

        private GameObjectWrapperMode mode = GameObjectWrapperMode.NONE;
        private int id;
        private int playerId;
        private bool isNetworkObject;
        private static Random rand = new System.Random((int)System.DateTime.Now.Ticks);
        private bool isPlayerCharacter;
        private long lastSerTimeStamp;
        private int typeId;

        public GameObjectWrapperDefault()
        {
            id = GetHashCode();
            playerId = -1;
            isNetworkObject = false;
            isPlayerCharacter = false;
            mode = GameObjectWrapperMode.NONE;
            PopulateAttrs();
        }

        public GameObjectWrapperDefault(bool isNetObj)
        {
            isNetworkObject = isNetObj;
            isPlayerCharacter = false;
            id = GetHashCode();
            playerId = -1;
            mode = GameObjectWrapperMode.NONE;
            PopulateAttrs();
        }

        public GameObjectWrapperDefault(bool isNetObj, int pId)
        {
            isNetworkObject = isNetObj;
            isPlayerCharacter = false;
            id = GetHashCode();
            playerId = pId;
            mode = GameObjectWrapperMode.NONE;
            PopulateAttrs();
        }

        public GameObjectWrapperDefault(bool isNetObj, int pId, bool isPcObj)
        {
            isNetworkObject = isNetObj;
            isPlayerCharacter = isPcObj;
            id = GetHashCode();
            playerId = pId;
            mode = GameObjectWrapperMode.NONE;
            PopulateAttrs();
        }

        public GameObjectWrapperDefault(GameObjectWrapper gow)
        {
            this.SetX(gow.GetX());
            this.SetY(gow.GetY());
            this.SetZ(gow.GetZ());
            this.SetId(gow.GetId());
            this.SetMode(gow.GetMode());
            this.SetRotW(gow.GetRotW());
            this.SetRotX(gow.GetRotX());
            this.SetRotY(gow.GetRotY());
            this.SetRotZ(gow.GetRotZ());
            this.SetTypeId(gow.GetTypeId());
            this.SetVelX(gow.GetVelX());
            this.SetVelY(gow.GetVelY());
            this.SetVelZ(gow.GetVelZ());
            this.SetPlayerId(gow.GetPlayerId());
            this.SetIsNetworkObject(gow.GetIsNetworkObject());
            this.SetIsPlayerCharacter(gow.GetIsPlayerCharacter());
            this.SetLastSerTimeStamp(gow.GetLastSerTimeStamp());
        }

        public virtual GameObjectWrapperSerMode GetObjSerMode()
        {
            return GameObjectWrapperDefault.OBJ_SER_MODE;
        }

        public virtual void SetObjSerMode(GameObjectWrapperSerMode osm)
        {
            GameObjectWrapperDefault.OBJ_SER_MODE = osm;
        }

        public virtual int[] GetBinaryLengths()
        {
            return GameObjectWrapperDefault.BINARY_LENGTHS;
        }

        public virtual void SetBinaryLengths(int[] bl)
        {
            GameObjectWrapperDefault.BINARY_LENGTHS = bl;
        }

        public virtual int[] GetBinaryOffsets()
        {
            return GameObjectWrapperDefault.BINARY_OFFSETS;
        }

        public virtual void SetBinaryOffsets(int[] bo)
        {
            GameObjectWrapperDefault.BINARY_OFFSETS = bo;
        }

        public virtual int GetTotalLength()
        {
            return GameObjectWrapperDefault.TOTAL_LENGTH;
        }

        public virtual void SetTotalLength(int len)
        {
            GameObjectWrapperDefault.TOTAL_LENGTH = len;
        }

        public virtual long GetLastSerTimeStamp()
        {
            return this.lastSerTimeStamp;
        }

        public virtual void SetLastSerTimeStamp(long l)
        {
            this.lastSerTimeStamp = l;
        }

        public virtual bool GetIsNetworkObject()
        {
            return this.isNetworkObject;
        }

        public virtual void SetIsNetworkObject(bool isNo)
        {
            this.isNetworkObject = isNo;
        }

        public virtual int GetMode()
        {
            return GameObjectWrapperDefault.Mode2Int(this.mode);
        }

        public virtual void SetMode(int m)
        {
            this.mode = GameObjectWrapperDefault.Int2Mode(m);
        }

        public virtual int GetPlayerId()
        {
            return this.playerId;
        }

        public virtual void SetPlayerId(int p)
        {
            this.playerId = p;
        }

        public virtual float GetX()
        {
            return this.x;
        }

        public virtual void SetX(float inX)
        {
            this.x = inX;
        }

        public virtual float GetY()
        {
            return this.y;
        }

        public virtual void SetY(float inY)
        {
            this.y = inY;
        }

        public virtual float GetZ()
        {
            return this.z;
        }

        public virtual void SetZ(float inZ)
        {
            this.z = inZ;
        }

        public virtual float GetRotX()
        {
            return this.rotX;
        }

        public virtual void SetRotX(float inX)
        {
            this.rotX = inX;
        }

        public virtual float GetRotY()
        {
            return this.rotY;
        }

        public virtual void SetRotY(float inY)
        {
            this.rotY = inY;
        }

        public virtual float GetRotZ()
        {
            return this.rotZ;
        }

        public virtual void SetRotZ(float inZ)
        {
            this.rotZ = inZ;
        }

        public virtual float GetRotW()
        {
            return this.rotW;
        }

        public virtual void SetRotW(float inW)
        {
            this.rotW = inW;
        }

        public virtual float GetVelX()
        {
            return this.velX;
        }

        public virtual void SetVelX(float velX)
        {
            this.velX = velX;
        }

        public virtual float GetVelY()
        {
            return this.velY;
        }

        public virtual void SetVelY(float velY)
        {
            this.velY = velY;
        }

        public virtual float GetVelZ()
        {
            return this.velZ;
        }

        public virtual void SetVelZ(float velZ)
        {
            this.velZ = velZ;
        }

        public virtual bool GetIsPlayerCharacter()
        {
            return this.isPlayerCharacter;
        }

        public virtual void SetIsPlayerCharacter(bool ipc)
        {
            this.isPlayerCharacter = ipc;
        }

        public virtual int GetId()
        {
            return this.id;
        }

        public virtual void SetId(int inId)
        {
            this.id = inId;
        }

        public virtual int GetTypeId()
        {
            return this.typeId;
        }

        public virtual void SetTypeId(int inTypeId)
        {
            this.typeId = inTypeId;
        }

        public virtual void PopulateAttrs()
        {
            if (Eds3Client.MOCK_MODE_ON)
            {
                if (GEN_RANDOM_ATTRS)
                {
                    SetX(rand.Next(short.MaxValue));
                    SetY(rand.Next(short.MaxValue));
                    SetZ(rand.Next(short.MaxValue));
                    SetRotX((float)rand.Next(360));
                    SetRotY((float)rand.Next(360));
                    SetRotZ((float)rand.Next(360));
                    SetRotW((float)rand.Next(360));
                    SetVelX((float)rand.Next(360));
                    SetVelY((float)rand.Next(360));
                    SetVelZ((float)rand.Next(360));
                    SetMode((int)(rand.Next(3)));
                    SetTypeId((int)(rand.Next(10)));
                }
                else
                {
                    SetX(0.0f);
                    SetY(0.0f);
                    SetZ(0.0f);
                    SetRotX(0.0f);
                    SetRotY(0.0f);
                    SetRotZ(0.0f);
                    SetRotW(0.0f);
                    SetVelX(0.0f);
                    SetVelY(0.0f);
                    SetVelZ(0.0f);
                    SetMode(GameObjectWrapperDefault.Mode2Int(GameObjectWrapperMode.NONE));
                    SetTypeId(0);
                }
            }
        }

        public virtual byte[] Serialize()
        {
            byte[] ret = new byte[TOTAL_LENGTH];
            this.lastSerTimeStamp = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();

            if (GameObjectWrapperDefault.OBJ_SER_MODE == GameObjectWrapperSerMode.L1)
            {
                //Utils.wr("Serialize L1");
                byte[] bX = Converter.toByte(this.GetX());
                byte[] bY = Converter.toByte(this.GetY());
                byte[] bZ = Converter.toByte(this.GetZ());
                byte[] bRotX = Converter.toByte(this.GetRotX());
                byte[] bRotY = Converter.toByte(this.GetRotY());
                byte[] bRotZ = Converter.toByte(this.GetRotZ());
                byte[] bRotW = Converter.toByte(this.GetRotW());
                byte[] bNid = Converter.toByte(this.GetId());
                byte[] bPid = Converter.toByte(this.GetPlayerId());
                byte[] bMode = Converter.toByte(this.GetMode());
                byte[] bType = Converter.toByte(this.GetTypeId());
                byte[] bLser = Converter.toByte(this.GetLastSerTimeStamp());

                System.Array.Copy(bX, 0, ret, BINARY_OFFSETS[0], BINARY_LENGTHS[0]);
                System.Array.Copy(bY, 0, ret, BINARY_OFFSETS[1], BINARY_LENGTHS[1]);
                System.Array.Copy(bZ, 0, ret, BINARY_OFFSETS[2], BINARY_LENGTHS[2]);
                System.Array.Copy(bRotX, 0, ret, BINARY_OFFSETS[3], BINARY_LENGTHS[3]);
                System.Array.Copy(bRotY, 0, ret, BINARY_OFFSETS[4], BINARY_LENGTHS[4]);
                System.Array.Copy(bRotZ, 0, ret, BINARY_OFFSETS[5], BINARY_LENGTHS[5]);
                System.Array.Copy(bRotW, 0, ret, BINARY_OFFSETS[6], BINARY_LENGTHS[6]);
                System.Array.Copy(bNid, 0, ret, BINARY_OFFSETS[7], BINARY_LENGTHS[7]);
                System.Array.Copy(bPid, 0, ret, BINARY_OFFSETS[8], BINARY_LENGTHS[8]);
                System.Array.Copy(bMode, 0, ret, BINARY_OFFSETS[9], BINARY_LENGTHS[9]);
                System.Array.Copy(bType, 0, ret, BINARY_OFFSETS[10], BINARY_LENGTHS[10]);
                System.Array.Copy(bLser, 0, ret, BINARY_OFFSETS[11], BINARY_LENGTHS[11]);
            }
            else
            {
                //Utils.wr("Serialize L2");
                byte[] bX = Converter.toByte(this.GetX());
                byte[] bY = Converter.toByte(this.GetY());
                byte[] bZ = Converter.toByte(this.GetZ());
                byte[] bRotX = Converter.toByte(this.GetRotX());
                byte[] bRotY = Converter.toByte(this.GetRotY());
                byte[] bRotZ = Converter.toByte(this.GetRotZ());
                byte[] bRotW = Converter.toByte(this.GetRotW());
                byte[] bVelX = Converter.toByte(this.GetVelX());
                byte[] bVelY = Converter.toByte(this.GetVelY());
                byte[] bVelZ = Converter.toByte(this.GetVelZ());
                byte[] bNid = Converter.toByte(this.GetId());
                byte[] bPid = Converter.toByte(this.GetPlayerId());
                byte[] bMode = Converter.toByte(this.GetMode());
                byte[] bType = Converter.toByte(this.GetTypeId());
                byte[] bLser = Converter.toByte(this.GetLastSerTimeStamp());

                System.Array.Copy(bX, 0, ret, BINARY_OFFSETS[0], BINARY_LENGTHS[0]);
                System.Array.Copy(bY, 0, ret, BINARY_OFFSETS[1], BINARY_LENGTHS[1]);
                System.Array.Copy(bZ, 0, ret, BINARY_OFFSETS[2], BINARY_LENGTHS[2]);
                System.Array.Copy(bRotX, 0, ret, BINARY_OFFSETS[3], BINARY_LENGTHS[3]);
                System.Array.Copy(bRotY, 0, ret, BINARY_OFFSETS[4], BINARY_LENGTHS[4]);
                System.Array.Copy(bRotZ, 0, ret, BINARY_OFFSETS[5], BINARY_LENGTHS[5]);
                System.Array.Copy(bRotW, 0, ret, BINARY_OFFSETS[6], BINARY_LENGTHS[6]);
                System.Array.Copy(bVelX, 0, ret, BINARY_OFFSETS[7], BINARY_LENGTHS[7]);
                System.Array.Copy(bVelY, 0, ret, BINARY_OFFSETS[8], BINARY_LENGTHS[8]);
                System.Array.Copy(bVelZ, 0, ret, BINARY_OFFSETS[9], BINARY_LENGTHS[9]);
                System.Array.Copy(bNid, 0, ret, BINARY_OFFSETS[10], BINARY_LENGTHS[10]);
                System.Array.Copy(bPid, 0, ret, BINARY_OFFSETS[11], BINARY_LENGTHS[11]);
                System.Array.Copy(bMode, 0, ret, BINARY_OFFSETS[12], BINARY_LENGTHS[12]);
                System.Array.Copy(bType, 0, ret, BINARY_OFFSETS[13], BINARY_LENGTHS[13]);
                System.Array.Copy(bLser, 0, ret, BINARY_OFFSETS[14], BINARY_LENGTHS[14]);
            }
            return ret;
        }

        public virtual void Deserialize(byte[] dIn)
        {
            if (GameObjectWrapperDefault.OBJ_SER_MODE == GameObjectWrapperSerMode.L1)
            {
                byte[] bX = Utils.PeelByteArray(dIn, BINARY_OFFSETS[0], BINARY_LENGTHS[0]);
                byte[] bY = Utils.PeelByteArray(dIn, BINARY_OFFSETS[1], BINARY_LENGTHS[1]);
                byte[] bZ = Utils.PeelByteArray(dIn, BINARY_OFFSETS[2], BINARY_LENGTHS[2]);
                byte[] bRotX = Utils.PeelByteArray(dIn, BINARY_OFFSETS[3], BINARY_LENGTHS[3]);
                byte[] bRotY = Utils.PeelByteArray(dIn, BINARY_OFFSETS[4], BINARY_LENGTHS[4]);
                byte[] bRotZ = Utils.PeelByteArray(dIn, BINARY_OFFSETS[5], BINARY_LENGTHS[5]);
                byte[] bRotW = Utils.PeelByteArray(dIn, BINARY_OFFSETS[6], BINARY_LENGTHS[6]);
                byte[] bNid = Utils.PeelByteArray(dIn, BINARY_OFFSETS[7], BINARY_LENGTHS[7]);
                byte[] bPid = Utils.PeelByteArray(dIn, BINARY_OFFSETS[8], BINARY_LENGTHS[8]);
                byte[] bMode = Utils.PeelByteArray(dIn, BINARY_OFFSETS[9], BINARY_LENGTHS[9]);
                byte[] bType = Utils.PeelByteArray(dIn, BINARY_OFFSETS[10], BINARY_LENGTHS[10]);
                byte[] bLser = Utils.PeelByteArray(dIn, BINARY_OFFSETS[11], BINARY_LENGTHS[11]);

                this.SetX(Converter.toFloat(bX));
                this.SetY(Converter.toFloat(bY));
                this.SetZ(Converter.toFloat(bZ));
                this.SetRotX(Converter.toFloat(bRotX));
                this.SetRotY(Converter.toFloat(bRotY));
                this.SetRotZ(Converter.toFloat(bRotZ));
                this.SetRotW(Converter.toFloat(bRotW));
                this.SetId(Converter.toInt(bNid));
                this.SetPlayerId(Converter.toInt(bPid));
                this.SetMode(Converter.toInt(bMode));
                this.SetTypeId(Converter.toInt(bType));
                this.SetLastSerTimeStamp(Converter.toLong(bLser));
            }
            else
            {
                byte[] bX = Utils.PeelByteArray(dIn, BINARY_OFFSETS[0], BINARY_LENGTHS[0]);
                byte[] bY = Utils.PeelByteArray(dIn, BINARY_OFFSETS[1], BINARY_LENGTHS[1]);
                byte[] bZ = Utils.PeelByteArray(dIn, BINARY_OFFSETS[2], BINARY_LENGTHS[2]);
                byte[] bRotX = Utils.PeelByteArray(dIn, BINARY_OFFSETS[3], BINARY_LENGTHS[3]);
                byte[] bRotY = Utils.PeelByteArray(dIn, BINARY_OFFSETS[4], BINARY_LENGTHS[4]);
                byte[] bRotZ = Utils.PeelByteArray(dIn, BINARY_OFFSETS[5], BINARY_LENGTHS[5]);
                byte[] bRotW = Utils.PeelByteArray(dIn, BINARY_OFFSETS[6], BINARY_LENGTHS[6]);
                byte[] bVelX = Utils.PeelByteArray(dIn, BINARY_OFFSETS[7], BINARY_LENGTHS[7]);
                byte[] bVelY = Utils.PeelByteArray(dIn, BINARY_OFFSETS[8], BINARY_LENGTHS[8]);
                byte[] bVelZ = Utils.PeelByteArray(dIn, BINARY_OFFSETS[9], BINARY_LENGTHS[9]);
                byte[] bNid = Utils.PeelByteArray(dIn, BINARY_OFFSETS[10], BINARY_LENGTHS[10]);
                byte[] bPid = Utils.PeelByteArray(dIn, BINARY_OFFSETS[11], BINARY_LENGTHS[11]);
                byte[] bMode = Utils.PeelByteArray(dIn, BINARY_OFFSETS[12], BINARY_LENGTHS[12]);
                byte[] bType = Utils.PeelByteArray(dIn, BINARY_OFFSETS[13], BINARY_LENGTHS[13]);
                byte[] bLser = Utils.PeelByteArray(dIn, BINARY_OFFSETS[14], BINARY_LENGTHS[14]);

                this.SetX(Converter.toFloat(bX));
                this.SetY(Converter.toFloat(bY));
                this.SetZ(Converter.toFloat(bZ));
                this.SetRotX(Converter.toFloat(bRotX));
                this.SetRotY(Converter.toFloat(bRotY));
                this.SetRotZ(Converter.toFloat(bRotZ));
                this.SetRotW(Converter.toFloat(bRotW));
                this.SetVelX(Converter.toFloat(bVelX));
                this.SetVelY(Converter.toFloat(bVelY));
                this.SetVelZ(Converter.toFloat(bVelZ));
                this.SetId(Converter.toInt(bNid));
                this.SetPlayerId(Converter.toInt(bPid));
                this.SetMode(Converter.toInt(bMode));
                this.SetTypeId(Converter.toInt(bType));
                this.SetLastSerTimeStamp(Converter.toLong(bLser));
            }
        }

        public virtual bool Equals(GameObjectWrapper gow)
        {
            if (gow != null)
            {
                if (Utils.VERBOSE)
                {
                    Utils.wr("Comparing key " + gow.GetId() + " to " + this.GetId());
                }

                if (GameObjectWrapperDefault.OBJ_SER_MODE == GameObjectWrapperSerMode.L1)
                {
                    if (gow.GetId() == this.GetId()
                        && (gow.GetX() - this.GetX()) < Utils.ZERO
                        && (gow.GetY() - this.GetY()) < Utils.ZERO
                        && (gow.GetZ() - this.GetZ()) < Utils.ZERO
                        && (gow.GetRotX() - this.GetRotX()) < Utils.ZERO
                        && (gow.GetRotY() - this.GetRotY()) < Utils.ZERO
                        && (gow.GetRotZ() - this.GetRotZ()) < Utils.ZERO
                        && (gow.GetRotW() - this.GetRotW()) < Utils.ZERO
                        && gow.GetPlayerId() == this.GetPlayerId()
                        && gow.GetMode() == this.GetMode()
                        && gow.GetType() == this.GetType()
                    )
                    {
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    if (gow.GetId() == this.GetId()
                        && (gow.GetX() - this.GetX()) < Utils.ZERO
                        && (gow.GetY() - this.GetY()) < Utils.ZERO
                        && (gow.GetZ() - this.GetZ()) < Utils.ZERO
                        && (gow.GetRotX() - this.GetRotX()) < Utils.ZERO
                        && (gow.GetRotY() - this.GetRotY()) < Utils.ZERO
                        && (gow.GetRotZ() - this.GetRotZ()) < Utils.ZERO
                        && (gow.GetRotW() - this.GetRotW()) < Utils.ZERO
                        && (gow.GetVelX() - this.GetVelX()) < Utils.ZERO
                        && (gow.GetVelY() - this.GetVelY()) < Utils.ZERO
                        && (gow.GetVelZ() - this.GetVelZ()) < Utils.ZERO
                        && gow.GetPlayerId() == this.GetPlayerId()
                        && gow.GetMode() == this.GetMode()
                        && gow.GetType() == this.GetType()
                    )
                    {
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
            }
            else
            {
                return false;
            }
        }

        public static int Mode2Int(GameObjectWrapperMode m)
        {
            if (m == GameObjectWrapperMode.UPDATE)
            {
                return 0;
            }
            else if (m == GameObjectWrapperMode.CREATE)
            {
                return 1;
            }
            else if (m == GameObjectWrapperMode.DESTROY)
            {
                return 2;
            }
            else if (m == GameObjectWrapperMode.NONE)
            {
                return 3;
            }
            else if (m == GameObjectWrapperMode.ERROR)
            {
                return 4;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_1)
            {
                return 5;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_2)
            {
                return 6;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_3)
            {
                return 7;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_4)
            {
                return 8;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_5)
            {
                return 9;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_6)
            {
                return 10;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_7)
            {
                return 11;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_8)
            {
                return 12;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_9)
            {
                return 13;
            }
            else if (m == GameObjectWrapperMode.SPECIAL_10)
            {
                return 14;
            }
            else if (m == GameObjectWrapperMode.ACTIVATE)
            {
                return 15;
            }
            else if (m == GameObjectWrapperMode.DEACTIVATE)
            {
                return 16;
            }
            else if (m == GameObjectWrapperMode.UPDATE_POS)
            {
                return 17;
            }
            else if (m == GameObjectWrapperMode.UPDATE_ROT)
            {
                return 18;
            }
            else if (m == GameObjectWrapperMode.UPDATE_VEL)
            {
                return 19;
            }
            else if (m == GameObjectWrapperMode.IGNORE)
            {
                return 20;
            }
            else if (m == GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR)
            {
                return 21;
            }
            else
            {
                return 4;
            }
        }

        public static GameObjectWrapperMode Int2Mode(int m)
        {
            if (m == 0)
            {
                return GameObjectWrapperMode.UPDATE;
            }
            else if (m == 1)
            {
                return GameObjectWrapperMode.CREATE;
            }
            else if (m == 2)
            {
                return GameObjectWrapperMode.DESTROY;
            }
            else if (m == 3)
            {
                return GameObjectWrapperMode.NONE;
            }
            else if (m == 4)
            {
                return GameObjectWrapperMode.ERROR;
            }
            else if (m == 5)
            {
                return GameObjectWrapperMode.SPECIAL_1;
            }
            else if (m == 6)
            {
                return GameObjectWrapperMode.SPECIAL_2;
            }
            else if (m == 7)
            {
                return GameObjectWrapperMode.SPECIAL_3;
            }
            else if (m == 8)
            {
                return GameObjectWrapperMode.SPECIAL_4;
            }
            else if (m == 9)
            {
                return GameObjectWrapperMode.SPECIAL_5;
            }
            else if (m == 10)
            {
                return GameObjectWrapperMode.SPECIAL_6;
            }
            else if (m == 11)
            {
                return GameObjectWrapperMode.SPECIAL_7;
            }
            else if (m == 12)
            {
                return GameObjectWrapperMode.SPECIAL_8;
            }
            else if (m == 13)
            {
                return GameObjectWrapperMode.SPECIAL_9;
            }
            else if (m == 14)
            {
                return GameObjectWrapperMode.SPECIAL_10;
            }
            else if (m == 15)
            {
                return GameObjectWrapperMode.ACTIVATE;
            }
            else if (m == 16)
            {
                return GameObjectWrapperMode.DEACTIVATE;
            }
            else if (m == 17)
            {
                return GameObjectWrapperMode.UPDATE_POS;
            }
            else if (m == 18)
            {
                return GameObjectWrapperMode.UPDATE_ROT;
            }
            else if (m == 19)
            {
                return GameObjectWrapperMode.UPDATE_VEL;
            }
            else if (m == 20)
            {
                return GameObjectWrapperMode.IGNORE;
            }
            else if (m == 21)
            {
                return GameObjectWrapperMode.DESTROY_INACTIVE_PLAYER_CHAR;
            }
            else
            {
                return GameObjectWrapperMode.ERROR;
            }
        }
    }
}
