﻿using System;
using System.Collections.Generic;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3
{
    public class GameRoom
    {
        public enum GameRoomMode
        {
            NO_STATE,
            IN_LOBBY,
            IN_GAME,
            STARTING_GAME,
            STOPPING_GAME,
            SPECIAL_1,
            SPECIAL_2,
            SPECIAL_3,
            SPECIAL_4,
            SPECIAL_5,
            SPECIAL_6,
            SPECIAL_7,
            SPECIAL_8,
            SPECIAL_9,
            SPECIAL_10,
            SERVER_SHUTTING_DOWN
        }

        public static double PLAYER_THRESHOLD_LAG_PRCT = 0.25;
        public static long PLAYER_THRESHOLD_INACTIVE_COMM_MS = 10000;

        private int spawnId = 0;
        public int id = -1;
        public Dictionary<string, Player> players = null;
        public GameRoomMode mode;
        public int dunGenSeed = 0;
        public string serverVersion = null;
        public List<Player> lagPlayers = null;
        public List<Player> inactivePlayers = null;
        public double msgAvgTimeMs = 0.0;
        public int msgTotalCnt = 0;
        public long msgTotalXportTimeMs = 0;
        public List<int> freeSpawns = null;
        public Eds3Server eds3server = null;

        public GameRoom()
        {
            this.players = new Dictionary<string, Player>();
            this.lagPlayers = new List<Player>();
            this.inactivePlayers = new List<Player>();
            this.mode = GameRoomMode.IN_LOBBY;
            this.freeSpawns = new List<int>();
        }

        public int GetSpawnId()
        {
            int ret = 0;
            if (this.freeSpawns.Count > 0)
            {
                ret = this.freeSpawns[0];
                this.freeSpawns.RemoveAt(0);
            }
            else
            {
                this.spawnId++;
                ret = this.spawnId;
            }
            return ret;
        }

        public void UpdateGameRoomStats()
        {
            int cnt = 0;
            double avg = 0.0;
            long totalDiff = 0;
            Player p = null;
            foreach (string key in this.players.Keys)
            {
                p = this.players[key];
                cnt += p.msgTotalCnt;
                totalDiff += p.msgTotalXportTimeMs;
            }

            if (cnt > 0)
            {
                avg = (double)totalDiff / (double)cnt;
            }
            else
            {
                avg = 0.0;
            }

            this.msgAvgTimeMs = avg;
            this.msgTotalCnt = cnt;
            this.msgTotalXportTimeMs = totalDiff;
        }

        public void UpdateLaggedInactivePlayers()
        {
            Player p = null;
            long now = System.DateTimeOffset.UtcNow.ToUnixTimeMilliseconds();

            foreach (string key in this.players.Keys)
            {
                p = this.players[key];
                double prctDiff = (p.msgAvgTimeMs - this.msgAvgTimeMs) / this.msgAvgTimeMs;

                Utils.wr("Inactive Player Time Calc: " + (now - p.lastCommTimeMs));
                if (now - p.lastCommTimeMs > PLAYER_THRESHOLD_INACTIVE_COMM_MS)
                {
                    p.isInactive = true;
                    p.isLagging = false;

                    if (this.inactivePlayers.Contains(p) != true)
                    {
                        this.inactivePlayers.Add(p);
                    }
                    else if (this.inactivePlayers.Contains(p) && (now - p.lastCommTimeMs) > PLAYER_THRESHOLD_INACTIVE_COMM_MS * 2)
                    {
                        Utils.wr("Drop from game...");
                        if(this.eds3server.GetPlayers().ContainsKey(p.ip))
                        {
                            this.eds3server.GetPlayers().Remove(p.ip);
                        }
                        this.inactivePlayers.Remove(p);
                        this.players.Remove(p.ip);
                        if (!freeSpawns.Contains(p.spawnId))
                        {
                            freeSpawns.Add(p.spawnId);
                        }
                    }

                    if (this.lagPlayers.Contains(p))
                    {
                        this.lagPlayers.Remove(p);
                    }
                }
                else if (prctDiff > 0 && prctDiff > PLAYER_THRESHOLD_LAG_PRCT)
                {
                    p.isLagging = true;
                    p.isInactive = false;

                    if (this.lagPlayers.Contains(p) != true)
                    {
                        this.lagPlayers.Add(p);
                    }

                    if (this.inactivePlayers.Contains(p))
                    {
                        this.inactivePlayers.Remove(p);
                    }
                }
                else
                {
                    p.isInactive = false;
                    p.isLagging = false;
                }
            }
        }

        public override string ToString()
        {
            string ret = "";
            ret += "CurrentSpawnId: " + this.spawnId + "\n";
            ret += "Id: " + this.id + "\n";
            ret += "PlayerCount: " + this.players.Count + "\n";
            ret += "CurrentMode: " + this.mode + "\n";
            ret += "DunGenSeed: " + this.dunGenSeed + "\n";
            ret += "ServerVersion: " + this.serverVersion + "\n";
            ret += "LagPlayerCount: " + this.lagPlayers.Count + "\n";
            ret += "MsgAvgTimeMs: " + this.msgAvgTimeMs + "\n";
            ret += "MsgTotalCnt: " + this.msgTotalCnt + "\n";
            ret += "MsgTotalXportTimeMs: " + this.msgTotalXportTimeMs + "\n";
            return ret;
        }
    }
}
