﻿using System;
using System.Threading;
using com.middlemind.Eds3.unity;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3.mock
{
    public class MockGenDungeon
    {
        private GameObjectManager man;
        private int dunGenSeed;
        private int playerId;              //also the spawnId
        public static bool SLEEP = false;
        public static bool RANDOM = false;
        private static Random rand = new System.Random((int)System.DateTime.Now.Ticks);
        public static bool MOCK_PLAYER_CHAR = true;

        public MockGenDungeon(int seed)
        {
            this.dunGenSeed = seed;
            this.playerId = 0;
            this.man = new GameObjectManagerDefault();
        }

        public MockGenDungeon(int seed, int spawnId)
        {
            this.dunGenSeed = seed;
            this.playerId = spawnId;
            this.man = new GameObjectManagerDefault();
        }

        public void GenerateObjects()
        {
            int min = 30;
            int max = 60;
            int objs = RANDOM ? rand.Next(min, max) : min;
            int nObjs = RANDOM ? rand.Next(min, max) : min;
            int i = 0;

            Utils.wr("Generating " + objs + " game objects.");
            for (i = 0; i < objs; i++)
            {
                this.man.AddGameObject(new GameObjectWrapperDefault(false, playerId), false);
            }

            Utils.wr("Generating " + nObjs + " network game objects.");
            for (i = 0; i < nObjs; i++)
            {
                this.man.AddGameObject(new GameObjectWrapperDefault(true, playerId), true);
            }

            if (MOCK_PLAYER_CHAR)
            {
                this.man.SetPlayerChar(new GameObjectWrapperDefault(true, playerId));
            }
        
            if (SLEEP)
            {
                try
                {
                    Thread.Sleep(5000);

                }
                catch (Exception e)
                {
                    Utils.wr(e.StackTrace);
                }
            }
        }

        public void GenerateSpawnObjects()
        {
            int min = 30;
            int max = 60;
            int nObjs = RANDOM ? rand.Next(min, max) : min;
            int i = 0;

            Utils.wr("Generating spawn " + nObjs + " spawn objects.");
            for (i = 0; i < nObjs; i++)
            {
                this.man.AddGameObject(new GameObjectWrapperDefault(true), true);
            }

            if (SLEEP)
            {
                try
                {
                    Thread.Sleep(5000);

                }
                catch (Exception e)
                {
                    Utils.wr(e.StackTrace);
                }
            }
        }

        public void RandomlyAdjustObjects()
        {
            int min = 10;
            int max = 20;
            int objs = (int)rand.Next(min, max);
            int nObjs = (int)rand.Next(min, max);
            int i = 0;

            Utils.wr("Adjusting " + objs + " game objects.");
            for (i = 0; i < objs; i++)
            {
                this.man.GetGameObjects().Remove(Utils.GetRandomKey(this.man.GetGameObjects()));
                this.man.AddGameObject(new GameObjectWrapperDefault(false), false);
            }

            Utils.wr("Adjusting " + nObjs + " network game objects.");
            for (i = 0; i < nObjs; i++)
            {
                this.man.GetNetworkGameObjects().Remove(Utils.GetRandomKey(this.man.GetNetworkGameObjects()));
                this.man.AddGameObject(new GameObjectWrapperDefault(true), true);
            }

            Utils.wr("Found " + this.man.GetGameObjects().Count + " gameObjects.");
            Utils.wr("Found " + this.man.GetNetworkGameObjects().Count + " networkGameObjects.");
        }

        public byte[] SerializeNetworkGameObjects()
        {
            return this.man.SerializeNetworkGameObjects();
        }

        public void DeserializeNetworkGameObjects(byte[] bIn)
        {
            this.man.DeserializeNetworkGameObjects(bIn);
        }

        public byte[] SerializeGameObjects()
        {
            return this.man.SerializeGameObjects();
        }

        public void DeserializeGameObjects(byte[] bIn)
        {
            this.man.DeserializeGameObjects(bIn);
        }

        public bool NetworkGameObjectsEquals(MockGenDungeon mgd)
        {
            if (mgd != null)
            {
                if (mgd.dunGenSeed == this.dunGenSeed && this.man.NetworkGameObjectsEquals(mgd.man))
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return false;
            }
        }

        public bool GameObjectsEquals(MockGenDungeon mgd)
        {
            if (mgd != null)
            {
                if (mgd.dunGenSeed == this.dunGenSeed && this.man.GameObjectsEquals(mgd.man))
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return false;
            }
        }

        public int GetGameObjectsCount()
        {
            return this.man.GetGameObjects().Count;
        }

        public int GetNetworkGameObjectsCount()
        {
            return this.man.GetNetworkGameObjects().Count;
        }

        public GameObjectManager GetGameObjectManager()
        {
            return this.man;
        }

        public GameObjectWrapper GetPlayerChar()
        {
            return this.man.GetPlayerChar();
        }
    }
}
