﻿using System;
using System.Collections.Generic;
using System.Net;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3
{
    public class Player : IComparable<Player>
    {
        public const int MAX_MSG_STATS = 100;
        public int id = -1;
        public int roomId = -1;
        public string ip = null;
        public int spawnId = -1;
        public int port = -1;
        public IPAddress ipAddress = null;
        public byte[] objPublish = null;
        public string clientVersion = null;
        public List<MsgStats> msgStats = null;
        public double msgAvgTimeMs = 0;
        public int msgTotalCnt = 0;
        public long msgTotalXportTimeMs = 0;
        public long lastCommTimeMs = 0;
        public bool isInactive = false;
        public bool isLagging = false;
        public float x;
        public float y;
        public float z;

        public Player(string Ip)
        {
            this.x = 0.0f;
            this.y = 0.0f;
            this.z = 0.0f;
            this.msgStats = new List<MsgStats>();
            this.ip = Ip;
        }

        public Player(int Id, int RoomId, string Ip)
        {
            this.x = 0.0f;
            this.y = 0.0f;
            this.z = 0.0f;
            this.msgStats = new List<MsgStats>();
            this.id = Id;
            this.roomId = RoomId;
            this.ip = Ip;
        }

        public Player(int Id, int RoomId, string Ip, int Port)
        {
            this.x = 0.0f;
            this.y = 0.0f;
            this.z = 0.0f;
            this.msgStats = new List<MsgStats>();
            this.id = Id;
            this.roomId = RoomId;
            this.ip = Ip;
            this.port = Port;
        }

        public void LogMsgStats(MsgStats stats)
        {
            if (this.msgStats.Count > MAX_MSG_STATS)
            {
                //this.msgStats.RemoveAt(this.msgStats.Count - 1);
                this.msgStats.RemoveAt(0);
            }
            this.msgStats.Add(stats);
            UpdateMsgStats();
        }

        private void UpdateMsgStats()
        {
            int len = this.msgStats.Count;
            int cnt = 0;
            double avg = 0.0;
            long totalDiff = 0;
            for (int i = 0; i < len; i++)
            {
                cnt++;
                totalDiff += this.msgStats[i].diff;
            }

            if (cnt > 0)
            {
                avg = (double)totalDiff / (double)cnt;
            }
            else
            {
                avg = 0.0;
            }

            this.msgAvgTimeMs = avg;
            this.msgTotalCnt = cnt;
            this.msgTotalXportTimeMs = totalDiff;
        }

        public override string ToString()
        {
            string ret = "";
            ret += "Id: " + this.id + "\n";
            ret += "RoomId: " + this.roomId + "\n";
            ret += "IP: " + this.ip + "\n";
            ret += "SpawnId/PlayerId: " + this.spawnId + "\n";
            ret += "Port: " + this.port + "\n";
            ret += "ClientVersion: " + this.clientVersion + "\n";
            ret += "MsgStatsCount: " + this.msgStats.Count + "\n";
            ret += "MsgAvgTimeMs: " + this.msgAvgTimeMs + "\n";
            ret += "MsgTotalCnt: " + this.msgTotalCnt + "\n";
            ret += "MsgTotalXportTimeMs: " + this.msgTotalXportTimeMs + "\n";
            return ret;
        }

        public int CompareTo(Player p)
        {
            Player p1 = this;
            Player p2 = p;

            int retVal = 0;
            if (p1.msgAvgTimeMs < p2.msgAvgTimeMs)
            {
                retVal = 1;
            }
            else if (p1.msgAvgTimeMs > p2.msgAvgTimeMs)
            {
                retVal = -1;
            }

            return retVal;
        }
    }
}
