﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using com.middlemind.Eds3.unity;

/*
 *
 * @author Victor G. Brusca, Middlemind Games
 */
namespace com.middlemind.Eds3
{
    public class Utils
    {
        public static float ZERO = 0.0001f;
        public static bool LOGGING_ON = true;
        public static bool VERBOSE = false;
        private static Random rand = new System.Random((int)System.DateTime.Now.Ticks);
        public static HandleDebugOut HANDLE_DEBUG_OUT = null;

        public static void wr(string s)
        {
            if(s == null) 
            {
                s = "";
            }

            if (LOGGING_ON)
            {
                if (HANDLE_DEBUG_OUT != null)
                {
                    HANDLE_DEBUG_OUT.wr(s);
                }
                else
                {
                    System.Console.Out.WriteLine(s);
                    Debug.WriteLine(s);
                }
            }
        }

        public static string PadRight(string s, int n)
        {
            return s.PadRight(s.Length + n, ' ');
        }

        public static string PadLeft(string s, int n)
        {
            return s.PadLeft(s.Length + n, ' ');
        }

        public static double SqrDist(GameObjectWrapper p1, Player p2)
        {
            double deltaX = p1.GetX() - p2.x;
            double deltaY = p1.GetY() - p2.y;
            double deltaZ = p1.GetZ() - p2.z;
            return (double)(deltaX * deltaX + deltaY * deltaY + deltaZ * deltaZ);
        }

        public static double SqrDist(double x, double y, double z, Player p2)
        {
            double deltaX = x - p2.x;
            double deltaY = y - p2.y;
            double deltaZ = z - p2.z;
            return (double)(deltaX * deltaX + deltaY * deltaY + deltaZ * deltaZ);
        }

        public static double Dist(GameObjectWrapper p1, Player p2)
        {
            double deltaX = p1.GetX() - p2.x;
            double deltaY = p1.GetY() - p2.y;
            double deltaZ = p1.GetZ() - p2.z;
            return Math.Sqrt((double)(deltaX * deltaX + deltaY * deltaY + deltaZ * deltaZ));
        }

        public static double Dist(double x, double y, double z, Player p2)
        {
            double deltaX = x - p2.x;
            double deltaY = y - p2.y;
            double deltaZ = z - p2.z;
            return Math.Sqrt((double)(deltaX * deltaX + deltaY * deltaY + deltaZ * deltaZ));
        }

        public static byte[] PeelByteArray(byte[] b, int start, int len)
        {
            byte[] nb = new byte[len];
            System.Array.Copy(b, start, nb, 0, len);
            return nb;
        }

        public static int[] GetOffsets(byte[][] p)
        {
            int[] ret = new int[p.Length + 1];
            int off = 0;
            int len = p.Length;

            for (int i = 0; i < len; i++)
            {
                ret[i] = off;
                off += p[i].Length;
            }

            ret[ret.Length - 1] = off;
            return ret;
        }

        public static int[] GetLengths(byte[][] p)
        {
            int[] ret = new int[p.Length];
            int len = p.Length;

            for (int i = 0; i < len; i++)
            {
                ret[i] = p[i].Length;
            }

            return ret;
        }

        public static int GetRandomIndex(List<GameObjectWrapper> list)
        {
            int len = list.Count;
            return rand.Next(len - 1);
        }

        public static int GetRandomKey(Dictionary<int, GameObjectWrapper> ht)
        {
            List<int> lst = new List<int>(ht.Keys);
            return lst[rand.Next(ht.Keys.Count)];
        }
    }
}
